/* $Id: ber.hxx,v 1.4 1997/09/28 16:28:25 cnidr Exp $ */
/************************************************************************
Copyright Notice

Copyright (c) MCNC, Clearinghouse for Networked Information Discovery and
Retrieval, 1994, 1995. 

Permission to use, copy, modify, distribute, and sell this software and
its documentation, in whole or in part, for any purpose is hereby granted
without fee, provided that

1. The above copyright notice and this permission notice appear in all
copies of the software and related documentation. Notices of copyright
and/or attribution which appear at the beginning of any file included in
this distribution must remain intact. 

2. Users of this software agree to make their best efforts (a) to return
to MCNC any improvements or extensions that they make, so that these may
be included in future releases; and (b) to inform MCNC/CNIDR of noteworthy
uses of this software. 

3. The names of MCNC and Clearinghouse for Networked Information Discovery
and Retrieval may not be used in any advertising or publicity relating to
the software without the specific, prior written permission of MCNC/CNIDR. 

THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY WARRANTY
OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE. 

IN NO EVENT SHALL MCNC/CNIDR BE LIABLE FOR ANY SPECIAL, INCIDENTAL,
INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES WHATSOEVER
RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT ADVISED OF THE
POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY, ARISING OUT OF OR
IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE. 
************************************************************************/
/*@@@
File:		ber.hxx
Version:	1.00
Description:	Wrapper classes for OCLC berutils, BER record code
Author:		Kevin Gamiel, kevin.gamiel@cnidr.org
Note:		Portions of this code and code design were provided by Mike 
		Gursky of Ovid, Inc.  Thanks, Mike!
@@@*/
#ifndef _BER_HXX_
#define _BER_HXX_

#include <iostream.h>

#include "gdt.h"
#include "berutil.hxx"
#include "string.hxx"


//***************************************************************************
//
//	class ASN1TAG - ASN1 Tag
//
//	MEMBERS:
//		Number		tag number
//		Class		tag class
//		ASN1TAG()	constructor
//		==		equivalence operator
//
//***************************************************************************
//@ManMemo: ASN.1 Tag  Class
/*@Doc:
\begin{description}
\item[Function] This class provides a simple base class for defining and manipulating ASN.1 tags as described in {\it The Open Book}.
\item[Algorithms]  None.  This is a container class.
\item[Flow] This is a fundamental c++ class.    Flow is described as needed within the individual method descriptions.
\item[Packaging] This fundamental class is linked and archived as determined by the compiling software.
\item[Security] This class has no security role or function.
\end{description}
*/
class ASN1TAG {
public:
  enum { 	DEFAULT_TAG = -1,
		DEFAULT_CLASS = ASN1_CONTEXT };
  //@ManMemo: BER Tag Value
  UINT	Number;	// tag number

  //@ManMemo: BER Class
  CHR	Class; // tag class

  //@ManMemo: Basic ASN.1 Tag Constructor
  ASN1TAG(INT theNum = DEFAULT_TAG, CHR theClass = DEFAULT_CLASS) {
    Number = theNum;
    Class = theClass;
  }
  
/*@Doc:  Creates an ASN.1 tag with assigned tag and class values.  
         The defaults are set in the system defines ({\tt defines.hxx}.*/
  
/// Assign a ASN.1 tag to another tag
  operator ==(const ASN1TAG& tag) const
    {
      return((Number == tag.Number) && (Class == tag.Class))?1:0;
    }
/*@Doc: Copy an existing tag into another tag.  Returns 1 on success, 
        0 on failure.*/
};

//***************************************************************************
//
//	class ASN1?TAG		- ASN1 Tag of specific class
//				- where ? is U,A,C,P for Universal, 
//                                Application,
//					Context and Private, respectively.
//
//	MEMBERS:
//		ASN1?Tag()	constructor
//
//***************************************************************************

/// ASN.1 Universal Tag Base Class
/*@Doc:
\begin{description}

\item[Function]  Derive a predefined Universal ASN.1 Tag
\item[Algorithms] Set the new ASN.1 tag to UNIVERSAL

\item[Flow] This is a fundamental c++ class.    Flow is described as needed within the individual method descriptions.
\item[Packaging] This fundamental class is linked and archived as determined by the compiling software.
\item[Security] This class has no security role or function.
\end{description}
*/

class ASN1TAGU
  : public ASN1TAG 
{
public:
    /// Universal Tag Constructor
    ASN1TAGU(UINT num) : ASN1TAG(num, ASN1_UNIVERSAL) {}
/*@Doc:  Creates an ASN.1 tag with default class  set to UNIVERSAL*/
};


/// ASN.1 Application Tag Base Class
/*@Doc:
\begin{description}

\item[Function]  Derive a predefined Application ASN.1 Tag
\item[Algorithms] Set the new ASN.1 tag to Application

\item[Flow] This is a fundamental c++ class.    Flow is described as needed within the individual method descriptions.
\item[Packaging] This fundamental class is linked and archived as determined by the compiling software.
\item[Security] This class has no security role or function.
\end{description}
*/
class ASN1TAGA 
  : public ASN1TAG 
{
public:
  /// Application Tag Constructor
  ASN1TAGA(UINT num) : ASN1TAG(num, ASN1_APPLICATION) {}
/*@Doc:  Creates an ASN.1 tag with default class  set to APPLICATION*/
};


/// ASN.1 Context Tag Base Class
/*@Doc:
\begin{description}

\item[Function]  Derive a predefined Context ASN.1 Tag
\item[Algorithms] Set the new ASN.1 tag to Context

\item[Flow] This is a fundamental c++ class.    Flow is described as needed within the individual method descriptions.
\item[Packaging] This fundamental class is linked and archived as determined by the compiling software.
\item[Security] This class has no security role or function.
\end{description}
*/
class ASN1TAGC 
  : public ASN1TAG 
{
public:
  /// Context Tag Constructor
  ASN1TAGC(UINT num) : ASN1TAG(num, ASN1_CONTEXT) {}
/*@Doc:  Creates an ASN.1 tag with defaults class  set to CONTEXT*/
};


/// ASN.1 Private Tag Base Class
/*@Doc:
\begin{description}

\item[Function]  Derive a predefined Private ASN.1 Tag
\item[Algorithms] Set the new ASN.1 tag to Private

\item[Flow] This is a fundamental c++ class.    Flow is described as needed within the individual method descriptions.
\item[Packaging] This fundamental class is linked and archived as determined by the compiling software.
\item[Security] This class has no security role or function.
\end{description}
*/
class ASN1TAGP 
  : public ASN1TAG 
{
public:
  ///Private Tag Constructor
  ASN1TAGP(UINT num) : ASN1TAG(num, ASN1_PRIVATE) {}
/*@Doc:  Creates an ASN.1 tag with defaults class  set to PRIVATE*/
};


//***************************************************************************
//
//      class BERTREE - Base class for a single BER record. 
//
//***************************************************************************
/// Base Class for single BER Record
/*@Doc:
\begin{description}
\item[Function] This class is related to the blk\_dir data structure from
  the OCLC berutils.  In the berutil model, the blk\_dir
  acts as a memory manager for a tree of dir\_nodes.
  BERTREE contains the root of a BER tree, which points
  to the managing blk\_dir.  Through the friend class BERBROWSER,
  nodes are added, deleted and browsed.
\item[Algorithms]
\item[Flow] This is a fundamental c++ class.    Flow is described as needed within the individual method descriptions.
\item[Packaging] This fundamental class is linked and archived as determined by the compiling software.
\item[Security] This class has no security role or function.
\end{description}
*/

class BERTREE 
{
  friend class BERBROWSER;
protected:
///Initialize the BERTREE
  void Init();
/*@Doc:  Perform BERTREE initialization functions*/

  /// root of directory structure
  PDATA_DIR     c_dir;          
  /// default number of dir_nodes to allocate
  enum { NUMDIRS = 15 };        
  INT           c_exit_on_error,
                c_ptrlist_count,
                c_ptrlist_maxcount;
  UCHR        **c_ptrlist;

  UCHR	       *c_buf;

public:
/// Constructor - Create BERTREE from BER file.
  BERTREE(const STRING & Filename); 
/*@Doc:  Set up BERTREE from previously stored BERTREE file */

/// Constructor - Create BERTREE from buffered BER.
  BERTREE(UCHR *buf, INT4 len);
/*@Doc:  Set up a new BERTREE from a character buffer containing existing BER data*/

/// Constructor - Create a new tagged BERTREE 
  BERTREE(ASN1TAG tag);
/*@Doc: Create a new, empty BERTREE with a specified tag*/

/// Basic Destructor
  ~BERTREE();
/*@Doc:  Delete this BERTREE*/

/// Set Exit On Error flag
  void ExitOnError(INT YesOrNo) { c_exit_on_error = YesOrNo; }
/*@Doc:  Define whether or not a  graceless exit on error should be performed*/

/// Set Error Message
  void Error(CHR *Msg);
/*@Doc:  Set an error message to be displayed*/
  
  /// Add a new data pointer
  void AddPointer(UCHR *p);
/*@Doc:  Insert a new character data pointer into the BERTREE.  These 
  pointers  aren't cleaned up by the {\bf BERTREE} destructor - this 
  function is handled by the {\bf BERBROWSER} friend class if this 
  method was invoked by {\bf BERBROWSER}. */

  /// Delete data pointers
  void DeletePointers();
/*@Doc:  Delete any data pointers in this BERTREE*/
  
  /// Get a character record
  PUCHR GetRecord(INT4 *len) { return bld_rec(c_dir, len); }
/*@Doc:  Extract contents of tree record into a buffer.  This should not 
  be confused with a Z39.50 data record in the context of GPI.*/

  /// Get the length of a record
  void GetRecordLength(INT4 *Length) { *Length = rec_len(c_dir); }
/*@Doc: Return the length of the current tree record*/

  /// Hex Dump this tree to a file  
  void HexDir(ostream &os) { hex_dirs(c_dir, 0, os); }
/*@Doc:  Write this tree (in hex) to a file*/

  /// Hex Dump this tree to STDERR
  void HexDir() { HexDir(cerr); }
/*@Doc:  Write this tree (in hex) to the STDERR stream*/
  
/// Stream input function
  friend ostream & operator<<(ostream& os, BERTREE& b);
/*@Doc:  Load this tree from a stream*/
};
typedef BERTREE *PBERTREE;

//***************************************************************************
//
//	class BERBROWSER
//
//***************************************************************************
typedef BERBROWSER *PBERBROWSER;
/// Main BERTREE Memory Manager class

/*@Doc:
\begin{description}

\item[Function]   A BERBROWSER consists of a BERTREE as generated by the OCLC BER 
utilities, and a list of memory blocks, which are the allocated memory pointers
referred to by the BERTREE.  Since the OCLC BERUTILS don't copy or maintain
the data, only the pointers, we must manage this memory as a block of
accessible pointers.  Management occurs in the BERBROWSER so we can do
garbage collection when the browser is destroyed, and to make the BER management code re-entrant.
\item[Algorithms] Create a tree of memory blocks, each associated with the tag stored in the same level of the internal BERTREE.  As a tagged block of
information is added to the BERBROWSER, add the tag and pointer to the BERTREE
using the OCLC BER utilities, and maintain a separate pointer to the associated
memory block for freeing upon destruction of the object.

\item[Flow]  Most functions are handled by the BER utilities.  This class
keeps track of the memory pointers caontaing the data handled by the BERTREE.
\item[Packaging] This fundamental class is linked and archived as determined by the comiling software.
\item[Security] This class has no security role or function.
\end{description}
*/

class BERBROWSER {
private:
  BERTREE* c_block;
  DATA_DIR *c_dirptr;
  
/// Locate a DATA\_DIR with a given tag
  DATA_DIR *FindTag(ASN1TAG Tag);
/*@Doc:  Returns the DATA\_DIR identified by the given tag.*/

/// Dump a DATA\_DIR in HTML
  void HTMLDump(DATA_DIR *d);
/*@Doc:  Dump the specified internal DATA\_DIR in HTML*/

/// Print entire node in HTML
  void PrintHTMLNode(DATA_DIR *d);
/*@Doc:  Emits HTML representation of entire node*/

public:
  /// Build Browser from BERTREE
  BERBROWSER(BERTREE *Tree);
/*@Doc:  Basic constructor - associates BERTREE with BERBROWSER*/

  /// Construct a new Browser from existing BERBROWSER
  BERBROWSER(const BERBROWSER & b);
/*@Doc: Creates a new browser based on contents of old browser.  The old 
  browser is duplicated, but not affected*/

  /// Destructor - zap BERTREE and clean up memory list
  ~BERBROWSER() {}
/*@Doc:  Delete  BERTREE and clean up memory list*/

  /// Dump out browser in hex
  void HexDir() { c_block->HexDir(); }
/*@Doc:  Hexadecimal dump of BERBROWSER contents to stdout.  Generally 
  used for debugging*/

  /// Dump out browser in HTML
  void HTMLDump();
/*@Doc:  Similar to hexdir, but creates a ragged HTML dump.  This method 
  is incomplete, and useful only for debugging*/
  
  /// Get Directory Parent
  INT GetParent(BERBROWSER * Parent);
/*Doc:  Return a browser set to the parent BERTREE directory.  Returns 1 on success, 0 on failure.*/

  /// Get Directory by Tag Number
  INT GetSubdirectory(ASN1TAG Tag, BERBROWSER * Subdir);
/*@Doc:  Get specific BERTREE directory by ASN.1 tag number.  Returns 1 on success, 0 on failure. */

  /// Get Directory by Directory Number
  INT GetSubdirectoryX(INT4 DirNum, BERBROWSER * Subdir);
/*@Doc: Get specific BERTREE directory by position.  Returns 1 on success, 0 on failure. */

  /// Get Number of Subdirectories
  INT4 GetSubdirectoryCount() { return c_dirptr->count; }
/*@Doc: Get a count of subdirectories in this browser*/

  /// Add Subdirectory to a browser with generic tag
  INT AddSubdirectory(ASN1TAG Tag, BERBROWSER * Newdir);
/*@Doc:  Add a subdirectory to external browser with a ASN.1 Generic tag.  Returns 1 on success, 0 on failure. */

  /// Add Subdirectory to a browser with universal tag
  INT AddSubdirectoryU(ASN1TAGU Tag, BERBROWSER * Newdir);
/*@Doc:  Add a subdirectory to external browser with a ASN.1 Universal tag. Returns 1 on success, 0 on failure.  */

  /// Add Subdirectory to current browser with universal  tag
  INT AddSubdirectoryU(ASN1TAGU Tag);
/*@Doc:  Add a subdirectory to this browser with a Universal ASN.1 tag.  Returns 1 on success, 0 on failure. */

  /// Add Subdirectory to current browser with generic tag
  INT AddSubdirectory(ASN1TAG Tag);
/*@Doc:  Add a subdirectory to this browser with a ASN.1 Generic tag.   Returns 1 on success, 0 on failure.*/

  /// Test if current browser has a tag
  INT HasTag(ASN1TAG Tag) { return FindTag(Tag)?1:0; }
/*@Doc:  Return 1 if this browser has the specified tag, 0 otherwise*/

  /// Get the field ID of a directory with this tag in this browser
  void GetTag(ASN1TAG *Tag) { *Tag = c_dirptr->fldid; }
/*@Doc:  Set {\tt Tag} to current field ID (tag value) of this {\bf BERBROWSER}*/
  
  
  // Leaf operations
  
  /// Add a NULL leaf to this browser
  INT AddNULL(ASN1TAG Tag);
/*@Doc: Inserts NULL leaf in current browser. Returns 1 on success, 0 on failure.  */

  /// Check for existence of NULL leaf in this browser
  INT HasNULL(ASN1TAG Tag);
/*@Doc: Returns 1 if this browser contains a NULl leaf.*/
  
  /// Add a generic tag character string to this browser
  INT AddChar(ASN1TAG Tag, const STRING& Data);
/*@Doc:  Adds a character string with a generic tag to this browser.  Browser manages memory used by string. Returns 1 on success, 0 on failure.*/

  /// Add a Universal tag character string to this browser
  INT AddCharU(ASN1TAGU Tag, const STRING& Data);
/*@Doc:  Adds a character string with a universal tag to this browser.  Browser manages memory used by string. Returns 1 on success, 0 on failure.*/

  /// Get character string with a given tag from this browser
  INT GetChar(ASN1TAG Tag, STRING * Data);
/*@Doc:  Find and retrieve a character string with a given tag.  The string
is {\bf copied} into the STRING object pointed to by {\tt Data}. Returns 1 on success, 0 on failure.*/
    
  /// Add a bit string to this browser
  INT AddBitString(ASN1TAG Tag, const STRING& Data);
/*@Doc:  Adds and ASN.1 bitstring to the current browser.   Returns 1 on success, 0 on failure.*/

  /// Get a bit string from this browser
  INT GetBitString(ASN1TAG Tag, STRING *Data);
  /*@Doc:  Retrieve a bit string with a specific tag number from this browser. .  Returns 1 on success, 0 on failure.*/

  /// Get an Integer value from this browser
  INT GetInt(ASN1TAG Tag, INT4 *Data);
/*@Doc:  Retrieve a integer with a specific tag number from this browser.  Returns 1 on success, 0 on failure.*/

  /// Add an Integer value to this browser
  INT AddInt(ASN1TAG Tag, INT4 Data);
  /*@Doc:  Add an integer to this browser, tagged with a specific tag.  Returns 1 on success, 0 on failure.*/

  /// Add an OID to this browser
  INT AddOID(ASN1TAG Tag, const STRING& Data);
/*@Doc:  Add an Object Identifier string to this browser, tagged with a specific tag.  Returns 1 on success, 0 on failure.*/

  /// Get an OID from this browser
  INT GetOID(ASN1TAG Tag, STRING * Data);
/*@Doc:  Retrieve an Object Identifier with a particular tag from this browser.  Returns 1 on success, 0 on failure.*/

};

//
// Set of macros for making common method declarations easier
//
#define DD_HASTAG(Name,Tag)\
inline INT Has##Name() const\
{ return c_root->HasTag(Tag);};

#define DD_GET(Name,Tag,Type,Suffix)\
inline INT Get##Name(Type *argValue) const\
{ return c_root->Get##Suffix(Tag,argValue);};

#define DD_ADD(Name,Tag,Type,Suffix)\
inline INT Add##Name(const Type & argValue) const\
{ return c_root->Add##Suffix(Tag,argValue);};

#define DD_ADDONSTACK(Name,Tag,Type,Suffix)\
inline INT Add##Name(Type argValue) const\
{ return c_root->Add##Suffix(Tag,argValue);};

#define DD_ALL(Name,Tag,Type,Suffix)\
DD_HASTAG(Name,Tag)\
DD_GET(Name,Tag,Type,Suffix)\
DD_ADD(Name,Tag,Type,Suffix)
     
#define DD_ALLONSTACK(Name,Tag,Type,Suffix)\
DD_HASTAG(Name,Tag)\
DD_GET(Name,Tag,Type,Suffix)\
DD_ADDONSTACK(Name,Tag,Type,Suffix)
     
#define DD_BITSTRING(Name, Tag) DD_ALL(Name,(Tag),STRING,BitString)
#define DD_CHAR(Name, Tag) DD_ALL(Name,(Tag),STRING,Char)
#define DD_STRING(Name, Tag) DD_ALL(Name,(Tag),STRING,Char)
#define DD_INT(Name, Tag) DD_ALLONSTACK(Name,(Tag),INT4,Int)
#define DD_OID(Name, Tag) DD_ALLONSTACK(Name,(Tag),STRING,OID)
#define DD_BOOLEAN(Name, Tag) DD_ALLONSTACK(Name,(Tag),INT4,Int)
#define DD_SUBDIR(Name, Tag)\
DD_HASTAG(Name,Tag)\
inline INT Add##Name() const\
{ return c_root->AddSubdirectory(Tag);};\
inline INT Get##Name(PBERBROWSER & Dir) const\
{ return c_root->GetSubdirectory(Tag,Dir);};
#define DD_SUBDIRU(Name, Tag)\
DD_HASTAG(Name,Tag)\
inline INT Add##Name() const\
{ return c_root->AddSubdirectoryU(Tag);};\
inline INT Get##Name(PBERBROWSER & Dir) const\
{ return c_root->GetSubdirectory(Tag,Dir);};


#endif				// _BER_HXX_

