/* $Id: zclibase.hxx,v 1.5 1998/01/30 19:13:09 cnidr Exp $ */
/************************************************************************
Copyright Notice

Copyright (c) MCNC, Clearinghouse for Networked Information Discovery and
Retrieval, 1994. 

Permission to use, copy, modify, distribute, and sell this software and
its documentation, in whole or in part, for any purpose is hereby granted
without fee, provided that

1. The above copyright notice and this permission notice appear in all
copies of the software and related documentation. Notices of copyright
and/or attribution which appear at the beginning of any file included in
this distribution must remain intact. 

2. Users of this software agree to make their best efforts (a) to return
to MCNC any improvements or extensions that they make, so that these may
be included in future releases; and (b) to inform MCNC/CNIDR of noteworthy
uses of this software. 

3. The names of MCNC and Clearinghouse for Networked Information Discovery
and Retrieval may not be used in any advertising or publicity relating to
the software without the specific, prior written permission of MCNC/CNIDR. 

THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY WARRANTY
OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE. 

IN NO EVENT SHALL MCNC/CNIDR BE LIABLE FOR ANY SPECIAL, INCIDENTAL,
INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES WHATSOEVER
RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT ADVISED OF THE
POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY, ARISING OUT OF OR
IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE. 
************************************************************************/

/*@@@
File:		zclibase.hxx
Version:	1.00
Description:	Z39.50 client API
Author:		Kevin Gamiel, kevin.gamiel@cnidr.org
@@@*/

#ifndef _ZCLIENT_BASE_HXX_
#define _ZCLIENT_BASE_HXX_

#include <iostream.h>
#include "zsession.hxx"

#define ERR_INIT_DENIED 9998

/// Basic Z39.50 Client Class
/*@Doc:
\begin{description}

\item[Function] Performs all functions necessary to create and manage a Z39.50 client session.  This is based on the Z39.50 session manager class.  {\bf ZCLIBASE} is the high-level client session manager for GPI and should be used for all client development activities.
\item[Algorithms]  Various techniques and algorithms are used on a per-method basis to manage the client session.

\item[Flow] This is a fundamental c++ class.    Flow is described as needed within the individual method descriptions.  Note that all methods returning GDT\_BOOLEAN return GDT\_TRUE on success, GDT\_FALSE on failure.
\item[Packaging] This fundamental class is linked and archived as determined by the comiling software.
\item[Security] This class has no security role or function.
\end{description}
*/

class ZCLIENT_BASE 
  : public ZSESSION {
private:
    void StoreDefaults(const STRING & Group);
    void GetRecordsOrDiagnostics(BERBROWSER *b, 
				 ZRECORDLIST *Records);
	
    STRING	c_query,
		c_host,
		c_element_set_name,
		c_motd, c_general_message, c_error_msg;
    STRLIST	c_diaglist;
    INT 	c_error;
    INT4	c_hitcount;
    STRING	c_postings;

public:
	/// Basic Constructor
    ZCLIENT_BASE();
	/*@Doc:  Initialize the client session*/

	/// Basic Destructor
    virtual ~ZCLIENT_BASE();
	/*@Doc:  Perform C++ Cleanup operations*/

	/// Convert Error Code to String
    void MapError(INT Code, STRING *Message);
	/*@Doc:  Maps error code to human readable string. */
    //
    // All methods return GDT_TRUE on success, GDT_FALSE on failure.
    // Use GetLastError() to determine reason.
    //

    /// Initialize with Z39.50 Server
    GDT_BOOLEAN Initialize( const STRING &Host, const UINT Port);

	/// Initialize with Z39.50 Server
    GDT_BOOLEAN Initialize( const STRING &Host, const UINT Port, 
			    STRING pass);
/*@Doc: Initialize connection to given host on given port, using {\tt pass} as password.*/

/// Initialize with Z39.50 Server
    GDT_BOOLEAN Initialize( const STRING &Host, const UINT Port,
			    const STRING &GroupId,
			    const STRING &UserId,
			    const STRING &Password);
/*@Doc: Initialize connection to given host on given port, using {\tt Password} as password, and {\tt GroupId/UserId} as username.*/

/// Initialize with Z39.50 Server
    GDT_BOOLEAN Initialize( const STRING &Host, 
			    const UINT Port,
			    const STRING &RefId,
			    const STRING &ProtVer,
			    const STRING &Options,
			    const INT4 PrefMsgSize,
			    const INT4 ExceptionalRecSize,
			    const STRING &GroupId,
			    const STRING &UserId,
			    const STRING &Password,
			    const STRING &ImpId,
			    const STRING &ImpName,
			    const STRING &ImpVersion);	

/*@Doc:  Full-function access to Z39.50 Initialize Request PDU.  Programmer provides all parameters to generate Init Request*/

	
/// Perform a Search
    virtual GDT_BOOLEAN Search(const STRING &Database,
			       const STRING &Term,
			       const STRING &ResultSetName,
			       INT4 *HitCount,
			       STRING UserId, 
			       INT Start,
			       INT Fetch,
			       ZRECORDLIST *Records);
/*@Doc:  Search against given database with {\tt Term} and return results in {\tt Records}.  Return result formats, counts and items are set with the remaining parameters*/


/// Perform a Search
    virtual GDT_BOOLEAN Search(const STRING &Database, 
			       const STRING &Term,
			       const STRING &ResultSetName, 
			       const STRING& AttrSetId,
			       INT4 *HitCount,
			       STRING userId, 
			       INT Start, 
			       INT Fetch,
			       ZRECORDLIST *Records); 
/*@Doc:  Search against given database with {\tt Term} and return 
         results in {\tt Records}.  Return result formats, counts and 
         items are set with the remaining parameters */


///Request one or more records
    GDT_BOOLEAN Present( const INT4 Start,
			 const INT4 Count,
			 const STRING &PreferredRecordSyntax,
			 const STRING &ResultSetName,
			 const STRING &ElementSetName,
			 ZRECORDLIST *Records);

/*@Doc: Request one or more records from the indicated result set.*/

/// Return Message of the Day
    void GetMOTD(STRING *Msg) { *Msg = c_motd; }
/*@Doc:  Get the Message of the Day returned in the Init Response PDU.  This is CAS-specific*/

/// Return General Message
    void GetGeneralMessage(STRING *Msg) { *Msg = c_general_message; }
/*@Doc:  Return the CAS-specific General Message from the Init Response PDU*/

///Return Diagnostic Error List
    void GetDiagnosticList(STRLIST *MsgList) { *MsgList = c_diaglist; }
/*@Doc:  Return a list of diagnostic errors encountered*/

/// Perform a SCAN (EXPAND)
    GDT_BOOLEAN Scan(const CHR *DatabaseName,
		     const CHR *Term,
		     const INT PrefPos,
		     const INT StepSize,
		     const INT NumRequest,
		     ZRECORDLIST *Records);
    /*@Doc:  Issue a Z39.50 SCAN request and return SCAN results in 
      {\tt Records}.  For GPI, this emulates the Messenger EXPAND function*/

/// Perform a SCAN (EXPAND)
    GDT_BOOLEAN Scan(const CHR *DatabaseName,
		     const CHR *Term,
		     const CHR *AttributeSet,
		     const INT PrefPos,
		     const INT StepSize,
		     const INT NumRequest,
		     ZRECORDLIST *Records);
    /*@Doc:  Issue a Z39.50 SCAN request and return SCAN results in 
      {\tt Records}.  For GPI, this emulates the Messenger EXPAND function*/

///Issue a Z39.50 Sort Request
    GDT_BOOLEAN Sort(	const STRING &InputResultSetNames,
			const STRING &SortedResultSetName,
			const STRING &SortSequence);
/*@Doc:  Request server to sort records on given {\tt SortSequence} element set name.  Name the sorted results*/

    ///Get Hit Count
    INT4 GetHitCount() { return c_hitcount; }
/*@Doc:  return count of hits after a search*/

/// Get Last Error
    INT GetLastError() { return c_error; }
/*@Doc:  Return code of last error condition*/

    /// Request server to add new records to a database
    GDT_BOOLEAN Add_Records(const STRING &Database,
			    ZRECORDLIST *Records);
    /*@Doc: Issue a request to Z39.50 server to add a record to the specified
      database */

    /// Request server to delete records from a database
    GDT_BOOLEAN Delete_Records(const STRING &Database,
			       ZRECORDLIST *Records);
    /*@Doc: Issue a request to Z39.50 server to delete a record to the 
      specified database */

    /// Get the current error message
    void GetErrorMessage(STRING *Message);

};

#endif
