/* $Id: zpdu.cxx,v 1.13 1998/01/30 19:31:35 cnidr Exp $ */
/************************************************************************
Copyright Notice

Copyright (c) MCNC, Clearinghouse for Networked Information Discovery and
Retrieval, 1994. 

Permission to use, copy, modify, distribute, and sell this software and
its documentation, in whole or in part, for any purpose is hereby granted
without fee, provided that

1. The above copyright notice and this permission notice appear in all
copies of the software and related documentation. Notices of copyright
and/or attribution which appear at the beginning of any file included in
this distribution must remain intact. 

2. Users of this software agree to make their best efforts (a) to return
to MCNC any improvements or extensions that they make, so that these may
be included in future releases; and (b) to inform MCNC/CNIDR of noteworthy
uses of this software. 

3. The names of MCNC and Clearinghouse for Networked Information Discovery
and Retrieval may not be used in any advertising or publicity relating to
the software without the specific, prior written permission of MCNC/CNIDR. 

THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY WARRANTY
OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE. 

IN NO EVENT SHALL MCNC/CNIDR BE LIABLE FOR ANY SPECIAL, INCIDENTAL,
INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES WHATSOEVER
RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT ADVISED OF THE
POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY, ARISING OUT OF OR
IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE. 
************************************************************************/

/*@@@
File:		zpdu.cxx
Version:	1.00
Description:	Z39.50 PDU classes
Author:		Kevin Gamiel, kevin.gamiel@cnidr.org
@@@*/

#include <ctype.h>
#include <stdlib.h>
#include <string.h>

#if defined(MSDOS) || defined(_WIN32)
#include <strstrea.h>
#else
#include <strstream.h>
#endif

#include "defs.hxx"
#include "string.hxx"
#include "vlist.hxx"
#include "strlist.hxx"

#include "zpdu.hxx"
//#include "config.hxx"
#include "defines.hxx"

#ifndef IFSTR
#define IFSTR(a) if((a) && (*a))
#endif
#ifndef IFOSTR
#define IFOSTR(a) if(a.GetLength() > 0)
#endif

static GDT_BOOLEAN IsBoolean(const STRING& Term);
static GDT_BOOLEAN GetTerm(const STRING &Term, STRING &Term1, STRING &Term2);
static GDT_BOOLEAN GetOperator(const STRING &Term, STRING &TheOperator);
static GDT_BOOLEAN ParseList(const STRING &Term, STRLIST *AttrList, STRING *TermOnly);


//***************************************************************************
//
//	Init - Initializer common to all ZPDU class constructors.
//
//***************************************************************************
void 
ZPDU::Init()
{
  c_root = new BERBROWSER(this);
  c_browser = new BERBROWSER(this);
}


//***************************************************************************
//
//	ZPDU - Constructor for (decoding a) ZPDU object.
//
//		Buf - Buffer of data satisfying call to IsCompleteBER()
//		Length - length of Buf
//
//***************************************************************************
ZPDU::ZPDU(PUCHR Buf,INT4 Length) 
  : BERTREE(Buf,Length)
{
  Init();
}


//***************************************************************************
//
//	ZPDU - Constructor for ZPDU object.
//
//		Tag - valid ASN.1 Z39.50 PDU tag value
//
//***************************************************************************
ZPDU::ZPDU(ASN1TAG Tag) 
  : BERTREE(Tag)
{
  Init();
}


//***************************************************************************
//
//	~ZPDU - Destructor for ZPDU
//
//***************************************************************************
ZPDU::~ZPDU()
{
  if(c_root)
    delete c_root;
  if(c_browser)
    delete c_browser;
}


//***************************************************************************
//
//	ZINITBASE - Constructor for object shared by both Init Requests and
//		Init Responses.
//
//		Tag - either 20 or 21, Init Request or Init Response
//
//***************************************************************************
ZINITBASE::ZINITBASE(ASN1TAG Tag) 
  : ZPDU(Tag)
{
}


//***************************************************************************
//
//	ZINITBASE - Constructor for (decoding) object shared by both Init 
//		Requests and Init Responses.
//
//		Buf - Buffer of data satisfying call to IsCompleteBER()
//		Length - length of Buf
//
//***************************************************************************
ZINITBASE::ZINITBASE(PUCHR Buf, INT4 Length) 
  : ZPDU(Buf,Length)
{
}


//***************************************************************************
//
//	ZINITREQUEST - Constructor for (decoding) Init Request PDU
//
//		Buf - Buffer of data satisfying call to IsCompleteBER()
//		Length - length of Buf
//
//***************************************************************************
ZINITREQUEST::ZINITREQUEST(PUCHR Buf, INT4 Length) 
: ZINITBASE(Buf, Length)
{
}


//***************************************************************************
//
//	ZINITREQUEST - Constructor for Init Request PDU
//
//		RefId - Reference Identifier (NULL if none)
//		Options - Z39.50 options supported
//		PrefMsgSize - PreferredMessageSize
//		ExceptionalRecSize - Maximum record size supported
//		GroupId - Authentication group id (NULL if none)
//		UserId - Authentication user id (NULL if none)
//		Password - Authentication password (NULL if none)
//		ImpId - Implementation Identifier
//		ImpName - Implementation Name
//		ImpVersion - Implementation Version
//
//	NOTES:
//		Authentication
//			If GroupId, UserId and Password, idPass encoded.  
//			If UserId and Password, idPass encoded.
//			If Password only, open encoded.  
//				If Password only, should be of the form 
//				"username/password".  This encoding is not 
//				preferred.
//			If None, anonymous encoded
//
//		You can call this method passing PCHR instead of STRING
//			arguments.  In that case, the compiler will call 
//			STRING(PCHR x) for each PCHR arg and will call 
//			~STRING()at the end of the method.
//
//***************************************************************************
ZINITREQUEST::ZINITREQUEST(	const STRING RefId,
				const STRING ProtVer,
				const STRING Options,
				const INT4 PrefMsgSize,
				const INT4 ExceptionalRecSize,
				const STRING GroupId,
				const STRING UserId,
				const STRING Password,
				const STRING ImpId,
				const STRING ImpName,
				const STRING ImpVersion)
  :ZINITBASE(INITREQUEST_TAG)
{
  BERBROWSER *Browser;
	
  Browser = new BERBROWSER(this);

  AddReferenceId(RefId);
  AddProtocolVersion(ProtVer);
  AddOptions(Options);
  AddPreferredMessageSize(PrefMsgSize);
  AddExceptionalRecordSize(ExceptionalRecSize);

  //
  // Do we need to encode the Authentication?
  //
  IFOSTR(GroupId) {
    // Encode using idPass
    AddAuthentication();
    Browser->GetSubdirectory(AUTHENTICATION_TAG, Browser);
    Browser->AddChar(GROUPID_TAG, GroupId);
    Browser->AddChar(USERID_TAG, UserId);
    Browser->AddChar(PASSWORD_TAG, Password);
  } else IFOSTR(UserId) {
    // Encode using idPass
    AddAuthentication();
    Browser->GetSubdirectory(AUTHENTICATION_TAG, Browser);
    Browser->AddChar(USERID_TAG, UserId);
    Browser->AddChar(PASSWORD_TAG, Password);
  } else IFOSTR(Password) {
    // Encode using open
    AddAuthentication();
    Browser->GetSubdirectory(AUTHENTICATION_TAG, Browser);
    ASN1TAGU UTag(INTERNATIONALSTRING_TAG);
    Browser->AddChar(UTag, Password);
  }
  
  STRING IID,PW;
  STRING INAME;
  //  IID="UJ";
  //  PW="pto1/pto1";
  INAME="6,1,1,1";
  /* comment out next 3 lines for login access 
     AddAuthentication();
     Browser->GetSubdirectory(AUTHENTICATION_TAG, Browser);
     Browser->AddCharU(ASN1_GENERALSTRING, PW);
  */
  
  AddImplementationId(IID);	// was ImpId
  /*
    AddImplementationName(INAME);
    AddImplementationVersion(ImpVersion);
  */
  
  delete Browser;
}


GDT_BOOLEAN 
ZINITREQUEST::GetGroupId(STRING *GroupId)
{
  if(!c_root->HasTag(AUTHENTICATION_TAG)) 
    return GDT_FALSE;
  c_root->GetSubdirectory(AUTHENTICATION_TAG, c_browser);
  if(!c_browser->HasTag(GROUPID_TAG))
    return GDT_FALSE;
  c_browser->GetChar(GROUPID_TAG, GroupId);
  return GDT_TRUE;
}


GDT_BOOLEAN 
ZINITREQUEST::GetUserId(STRING *UserId)
{
  if(!c_root->HasTag(AUTHENTICATION_TAG)) 
    return GDT_FALSE;
  c_root->GetSubdirectory(AUTHENTICATION_TAG, c_browser);
  if(!c_browser->HasTag(USERID_TAG))
    return GDT_FALSE;
  c_browser->GetChar(USERID_TAG, UserId);
  return GDT_TRUE;
}


GDT_BOOLEAN 
ZINITREQUEST::GetPassword(STRING *Password)
{
  if(!c_root->HasTag(AUTHENTICATION_TAG)) 
    return GDT_FALSE;
  c_root->GetSubdirectory(AUTHENTICATION_TAG, c_browser);
  if(!c_browser->HasTag(PASSWORD_TAG))
    return GDT_FALSE;
  c_browser->GetChar(PASSWORD_TAG, Password);
  return GDT_TRUE;
}


GDT_BOOLEAN 
ZINITREQUEST::GetAuthentication(STRING *Password)
{
  if(!c_root->HasTag(AUTHENTICATION_TAG)) 
    return GDT_FALSE;
  c_root->GetSubdirectory(AUTHENTICATION_TAG, c_browser);
  if(!c_browser->HasTag(INTERNATIONALSTRING_TAG))
    return GDT_FALSE;
  ASN1TAGU UTag(INTERNATIONALSTRING_TAG);
  c_browser->GetChar(UTag, Password);
  return GDT_TRUE;
}


//***************************************************************************
//
//	ZINITRESPONSE - Constructor for (decoding) Init Response PDU
//
//		Buf - Buffer of data satisfying call to IsCompleteBER()
//		Length - length of Buf
//
//***************************************************************************
ZINITRESPONSE::ZINITRESPONSE(PUCHR Buf, INT4 Length):ZINITBASE(Buf, Length)
{
}


//***************************************************************************
//
//	ZINITRESPONSE - Constructor for Init Response PDU
//
//		RefId - Reference Identifier
//		Options - Z39.50 options supported
//		PrefMsgSize - PreferredMessageSize
//		ExceptionalRecSize - Maximum record size supported
//		Result - Result of initialize request
//			0 - failure
//			1 - success	
//		ImpId - Implementation Identifier
//		ImpName - Implementation Name
//		ImpVersion - Implementation Version
//
//		You can call this method passing PCHR instead of STRING
//			arguments.  In that case, the compiler will call 
//			STRING(PCHR x) for each PCHR arg and will call 
//			~STRING()at the end of the method.
//
//***************************************************************************
ZINITRESPONSE::ZINITRESPONSE(	const STRING& RefId,
				const STRING& ProtVer,
				const STRING& Options,
				const INT4 PrefMsgSize,
				const INT4 ExceptionalRecSize,
				const INT4 Result,
				const STRING& ImpId,
				const STRING& ImpName,
				const STRING& ImpVersion)
  : ZINITBASE(INITRESPONSE_TAG)
{
  AddReferenceId(RefId);
  AddProtocolVersion(ProtVer);
  AddOptions(Options);
  AddPreferredMessageSize(PrefMsgSize);
  AddExceptionalRecordSize(ExceptionalRecSize);
  AddResult(Result);
  AddImplementationId(ImpId);
  AddImplementationName(ImpName);
  AddImplementationVersion(ImpVersion);
}


//***************************************************************************
//
//	ZSEARCHREQUEST - Constructor for Search Request PDU
//
//		RefId - Reference Identifier
//		SmallSetUpperBound
//		LargeSetLowerBound
//		MedSetPresentNum
//			These values define when and how many records should
//			be piggy-backed on the search response.  If you
//			do NOT want records piggy-backed, rather you want
//			to explicitly request them from the result set via
//			Present, set to 0,1,0 respectively.  See the 
//			Z39.50 standard for more details.
//		ResultSetName - String identifier for the result set on
//			the server side (in case they support multiple
//			result sets.
//		DatabaseNames - A comma-delimited list of database names
//			in which to query.
//		SmallElementSetName - Record syntax for piggy-backed records.
//		MediumElementSetName - Record syntax for piggy-backed records.
//		PrefRecSyntax - Record syntax for piggy-backed records.
//		Query - A query string of the form:
//			"dog"
//			"AND(dog,cat)"
//			"OR(dog.cat)"
//			"dog[1=21,2=3]"
//			"AND(dog[1=21],OR(cat[1=4],mouse[1,4]))"
//***************************************************************************
ZSEARCHREQUEST::ZSEARCHREQUEST( const STRING& RefId,
				const INT4 SmallSetUpperBound,
				const INT4 LargeSetLowerBound,
				const INT4 MedSetPresentNum,
				const INT4 ReplaceIndicator,
				const STRING& ResultSetName,
				const STRING& DatabaseNames,
				const CHR *SmallElementSetName,
				const CHR *MediumElementSetName,
				const STRING& PrefRecSyntax,
				const STRING& Query,
				const STRING& AttrSetId,
				const STRING& UserId,
				const INT QueryType) 
  : ZSEARCHBASE(SEARCHREQUEST_TAG)
{
  
  SetuserId(UserId);
  AddReferenceId(RefId);
  AddSmallSetUpperBound(SmallSetUpperBound);
  AddLargeSetLowerBound(LargeSetLowerBound);
  AddMediumSetPresentNumber(MedSetPresentNum);
  AddReplaceIndicator(ReplaceIndicator);
  AddResultSetName(ResultSetName);
  
  // Add database names
  AddDatabaseNames();
  STRLIST sl;
  STRING s;
  STRINGINDEX i;
  STRING *os;

  if (DatabaseNames.Search(',')) {
    // OK - they sent us a list of database names
    sl.Split(',',DatabaseNames);
    for (INT kk=1; kk<=sl.GetTotalEntries();kk++) {
      sl.GetEntry(kk,&s);
      if (c_root->GetSubdirectory(DATABASENAMES_TAG, c_browser)) {
	os = new STRING(s);
	c_browser->AddChar(DATABASENAME_TAG, *os);
	delete os;
      }
    }
  } else {
    // Got a single name - store it away by itself
    s = DatabaseNames;
    if(c_root->GetSubdirectory(DATABASENAMES_TAG, c_browser)) {
      os = new STRING(s);
      c_browser->AddChar(DATABASENAME_TAG, *os);
      delete os;
    }
  }

  /* Old version
  s = DatabaseNames;
  STRINGINDEX i = s.Search(',');
  if(i)
    s.EraseAfter(i-1);
  if(c_root->GetSubdirectory(DATABASENAMES_TAG, c_browser)) {
    STRING *os;
    os = new STRING(s);
    c_browser->AddChar(DATABASENAME_TAG, *os);
    delete os;
  }
  */
  IFSTR(SmallElementSetName) {
    AddSmallSetElementSetNames();
    if(c_root->GetSubdirectory(SMALLSETELEMENTSETNAMES_TAG, 
			       c_browser))
      c_browser->AddChar(GENERICELEMENTSETNAME_TAG,
			 SmallElementSetName);
  }
  IFSTR(MediumElementSetName) {
    AddMediumSetElementSetNames();
    if(c_root->GetSubdirectory(MEDIUMSETELEMENTSETNAMES_TAG, 
			       c_browser))
      c_browser->AddChar(GENERICELEMENTSETNAME_TAG,
			 MediumElementSetName);
  }
  AddPreferredRecordSyntax(PrefRecSyntax);
  
  // Query
  AddQuery();
  c_root->GetSubdirectory(QUERY_TAG, c_browser);
  /*
     to test type 0 query, uncomment these lines and comment out those below
     
     c_browser->AddSubdirectory(TYPE0_TAG, c_browser);
     ASN1TAGU tag(ASN1_GENERALSTRING);
     c_browser->AddChar(tag, Query);
     */
  if(QueryType == 101) {
    c_browser->AddSubdirectory(TYPE1_TAG, c_browser);
    c_browser->AddOID(ASN1TAGU(ATTRIBUTESETID_TAG), AttrSetId);
    
    
    AddTerm(*c_browser, Query);
  }
  /*
     c_browser->AddSubdirectory(TYPE1_TAG, c_browser);
     c_browser->AddOID(ASN1TAGU(ATTRIBUTESETID_TAG), BIB1_ATTRSET_OID);
     c_browser->AddSubdirectory(OP_TAG, c_browser);
     c_browser->AddSubdirectory(ATTRIBUTESPLUSTERM_TAG, c_browser);
     BERBROWSER sub(this);
     c_browser->AddSubdirectory(ATTRIBUTELIST_TAG, &sub);
     sub.AddSubdirectory(ASN1TAGU(ASN1_SEQUENCE), &sub);
     sub.AddInt(ATTRIBUTETYPE_TAG, 1);
     sub.AddInt(NUMERIC_TAG, 4);
     c_browser->AddChar(TERMGENERAL_TAG, Query);
     */
}


void 
ZSEARCHREQUEST::SetuserId(const STRING& user)
{
  user.GetCString(GlobalUserId,256);
}


void 
ZSEARCHREQUEST::GetDatabaseName(STRING *DBName)
{
  *DBName = "";
  if(!c_root->HasTag(DATABASENAMES_TAG)) 
    return;
  c_root->GetSubdirectory(DATABASENAMES_TAG, c_browser);
  c_browser->GetChar(DATABASENAME_TAG, DBName);
}


ZSCANREQUEST::ZSCANREQUEST(     const CHR *RefId,
				const CHR *DatabaseNames,
				const CHR *AttributeSet,
				const CHR *TermList,
				INT4 StepSize,
				INT4 NumTermsRequested,
				INT4 PrefPosInResponse)
  : ZSCANBASE(SCANREQUEST_TAG)
{
  AddReferenceId(RefId);
  // Add database names.  We only support one until we get a good
  // STRLIST method for spliting on commas
  
  STRING s;
  s = DatabaseNames;
  STRINGINDEX i = s.Search(',');
  if(i)
    s.EraseAfter(i-1);
  BERBROWSER sub(this);
  c_root->AddSubdirectory(SCANDATABASENAMES_TAG, &sub);
  sub.AddChar(DATABASENAME_TAG, s);
  /*
     AddScanDatabaseNames();
     if(c_root->GetSubdirectory(SCANDATABASENAMES_TAG, c_browser))
     c_browser->AddChar(DATABASENAME_TAG, s);
     */
  AddTermListAndStartPoint();
  c_root->GetSubdirectory(ATTRIBUTESPLUSTERM_TAG, c_browser);
  BERBROWSER sub2(this);
  c_browser->AddSubdirectory(ATTRIBUTELIST_TAG, &sub2);
  BERBROWSER sub3(sub2);
  INT     j,
  attrcount;
  STRLIST AttrList;
  STRING TermOnly;
  ParseList(TermList, &AttrList, &TermOnly);
  attrcount = AttrList.GetTotalEntries();
  INT4 Type, Value;
  STRING Temp;
  for(j=1;j <= attrcount;j+=2) {
    AttrList.GetEntry(j, &Temp);
    Type = Temp.GetInt();
    AttrList.GetEntry(j+1, &Temp);
    Value = Temp.GetInt();
    sub2.AddSubdirectory(ASN1TAGU(ASN1_SEQUENCE), &sub3);
    sub3.AddInt(ATTRIBUTETYPE_TAG, Type);
    sub3.AddInt(NUMERIC_TAG, Value);
  }
  c_browser->AddChar(TERMGENERAL_TAG, TermOnly);
  
  AddStepSize(StepSize);
  AddNumberOfTermsRequested(NumTermsRequested);
  AddPreferredPositionInResponse(PrefPosInResponse);
}


//***************************************************************************
//
//	ZSCANRESPONSE - Constructor for Scan Response PDU
//
//
//***************************************************************************

ZSCANRESPONSE::ZSCANRESPONSE(	const CHR *RefId,
				INT4 StepSize,
				INT4 ScanStatus, 
				INT4 NumberOfEntriesReturned, 
				INT4 PositionOfTerm,
				const CHR *ListEntries,
				const CHR *AttributeSet)
  : ZSCANBASE(SCANRESPONSE_TAG)
     
{
  AddRetStepSize(StepSize);
  AddScanStatus(ScanStatus);
  AddNumberOfEntriesReturned(NumberOfEntriesReturned);
  AddPositionOfTerm(PositionOfTerm);
  // AddListEntries;
  // AttributeSet;
}

ZCLOSE::ZCLOSE(const CHR *RefId, INT4 Reason) : ZPDU(CLOSE_TAG)
{
  AddReferenceId(RefId);
  AddCloseReason(Reason);
}


//***************************************************************************
//
//	ZSEARCHRESPONSE - Constructor for Search Response PDU
//
//		RefId - Reference Identifier
//		ResultCount - Number of hits
//
//***************************************************************************
ZSEARCHRESPONSE::ZSEARCHRESPONSE(const STRING& RefId,
				 const INT4 ResultCount,
				 const INT4 NumberOfRecordsReturned,
				 const INT4 NextResultSetPosition,
				 const INT4 SearchStatus,
				 const INT4 ResultSetStatus,
				 const INT4 PresentStatus,
				 ZRECORDLIST *Records)
  : ZSEARCHBASE(SEARCHRESPONSE_TAG)
{
  AddReferenceId(RefId);
  AddResultCount(ResultCount);
  AddNumberOfRecordsReturned(NumberOfRecordsReturned);
  AddNextResultSetPosition(NextResultSetPosition);
  AddSearchStatus(SearchStatus);
  if(SearchStatus == GDT_FALSE)
    AddResultSetStatus(ResultSetStatus);
  else
    AddPresentStatus(PresentStatus);
  if(Records)
    Records->AddBER(*c_browser);
}


//***************************************************************************
//
//	ZPRESENTREQUEST - Constructor for Present Request PDU
//
//		RefId - Reference Identifier
//
//***************************************************************************
ZPRESENTREQUEST::ZPRESENTREQUEST(const STRING& RefId,
				 const STRING& ResultSetId,
				 const INT4 ResultSetStartPoint,
				 const INT4 NumberOfRecordsRequested,
				 const STRING& ElementSetName,
				 const STRING& PreferredRecordSyntax)
  : ZPRESENTBASE(PRESENTREQUEST_TAG)
{
  AddReferenceId(RefId);
  AddResultSetId(ResultSetId);
  AddResultSetStartPoint(ResultSetStartPoint);
  AddNumberOfRecordsRequested(NumberOfRecordsRequested);
  
  // Add Generic Element Set Name
  BERBROWSER sub(this);
  c_root->AddSubdirectory(SIMPLE_TAG, &sub);
  sub.AddChar(GENERICELEMENTSETNAME_TAG, ElementSetName);
  
  AddPreferredRecordSyntax(PreferredRecordSyntax);
}


void 
ZPRESENTREQUEST::GetElementSetName(STRING *ElementSetName)
{
  BERBROWSER sub(this);
  if(c_root->HasTag(SIMPLE_TAG)) {
    c_root->GetSubdirectory(SIMPLE_TAG, &sub);
    sub.GetChar(GENERICELEMENTSETNAME_TAG, ElementSetName);
  }
}


//***************************************************************************
//
//	ZPRESENTRESPONSE - Constructor for Present Response PDU
//
//		RefId - Reference Identifier
//
//***************************************************************************
ZPRESENTRESPONSE::ZPRESENTRESPONSE(const STRING& RefId,
				   const INT4 NumberOfRecordsReturned,
				   const INT4 NextResultSetPosition,
				   const INT4 PresentStatus,
				   ZRECORDLIST *Records)
  : ZPRESENTBASE(PRESENTRESPONSE_TAG)
{
  AddReferenceId(RefId);
  AddNumberOfRecordsReturned(NumberOfRecordsReturned);
  AddNextResultSetPosition(NextResultSetPosition);
  AddPresentStatus(PresentStatus);
  Records->AddBER(*c_browser);
}


//***************************************************************************
//
//	ZESREQUEST - Constructor for Extended Service Request PDU
//
//		RefId - Reference Identifier
//		Function - Create, delete, or modify
//		PackageType - OID of the extended service to use
//		PackageName - Task package name
//		UserId - User id
//		RetentionTime - Length of time to retain this task package
//		Description - Description of this task package
//		WaitAction - Whether the target should (or may) include
//			the task package in the response.  Choices are:
//			"wait", "wait if possible", "do not wait", or
//			"do not send task package back in the response"
//***************************************************************************
ZESREQUEST::ZESREQUEST(	const STRING& RefId,
			const INT4 Function,
			const STRING& PackageType,
			const STRING& PackageName,
			const STRING& UserId,
			const INT4 RetentionTime,
			const STRING& Description,
			const INT4 WaitAction)
  : ZESBASE(EXTENDEDSERVICESREQUEST_TAG)
{
  AddReferenceId(RefId);
  AddFunction(Function);
  AddPackageType(PackageType);
  AddPackageName(PackageName);
  AddUserId(UserId);
  AddRetentionTime(RetentionTime);
  AddPermissions();
  AddDescription(Description);
  AddWaitAction(WaitAction);
}


//***************************************************************************
//
//	ZESRESPONSE - Constructor for Extended Service Response PDU
//
//		RefId - Reference Identifier
//		OperationStatus - Operation Status
//		DiagRecord - Diagnostic record
//		TaskPackage - Task Package record
//***************************************************************************
ZESRESPONSE::ZESRESPONSE(	const STRING& RefId,
				const INT4 OperationStatus,
				ZDEFAULTDIAGFORMAT *DiagRecord,
				ZESTASKPACKAGE *TaskPackage)
  : ZESBASE(EXTENDEDSERVICESRESPONSE_TAG)
{

  AddReferenceId(RefId);
  AddOperationStatus(OperationStatus);

  if ( DiagRecord ) {
    c_root->AddSubdirectory(ES_DIAGNOSTICS_TAG, c_browser);

    ASN1TAGU seqtag(ASN1_SEQUENCE);
    BERBROWSER sub(*c_browser);

    c_browser->AddSubdirectory(seqtag, &sub);
    DiagRecord->AddBER(sub);
  }

  if ( TaskPackage ) {
    c_root->AddSubdirectory(ES_TASKPACKAGE_TAG, c_browser);
    TaskPackage->AddBER(*c_browser);
  }
}


GDT_BOOLEAN 
ZESRESPONSE::GetTaskStatus(INT4 *TaskStatus)
{
  if( !c_root->GetSubdirectory(ES_TASKPACKAGE_TAG, c_browser) )
    return GDT_FALSE;

  ASN1TAGU exttag(ASN1_EXTERNAL);

  if( !c_browser->GetSubdirectory(exttag, c_browser) )
    return GDT_FALSE;

  ASN1TAGU oidtag(ASN1_OBJECTIDENTIFIER);
  STRING OID;

  if( !c_browser->GetOID(oidtag, &OID) || OID != EXTENDEDSERVICES_OID)
    return GDT_FALSE;

  if( !c_browser->GetInt(ES_TP_TASKSTATUS_TAG, TaskStatus) )
    return GDT_FALSE;

  return GDT_TRUE;
}


//***************************************************************************
//
//	ZESUPDATEREQUEST - Constructor for ES Update Request PDU
//
//		RefId - Reference Identifier
//		Action - Action to perform
//		DatabaseName - Database to update
//		Records - Records to update in the database
//***************************************************************************
ZESUPDATEREQUEST::ZESUPDATEREQUEST( const STRING& RefId,
				    const INT4 Action,
				    const STRING& DatabaseName,
				    ZRECORDLIST *Records) 
  : ZESREQUEST(RefId,
	       ES_CREATE_FUNCTION,
	       ES_UPDATE_OID,
	       "",
	       "",
	       0,
	       "",
	       ES_WAIT_WAITACTION)
{

  c_root->AddSubdirectory(ES_TASKSPECIFIC_TAG, c_browser);

  ASN1TAGU exttag(ASN1_EXTERNAL);

  c_browser->AddSubdirectory(exttag, c_browser);

  ASN1TAGU oidtag(ASN1_OBJECTIDENTIFIER);

  c_browser->AddOID(oidtag, ES_UPDATE_OID);

  c_browser->AddSubdirectory(ES_TS_ESREQUEST_TAG, c_browser);

  BERBROWSER sub(*c_browser);

  c_browser->AddSubdirectory(ES_TS_TOKEEP_TAG, &sub);

  sub.AddInt(ES_UPDATE_ACTION_TAG, Action);
  sub.AddChar(ES_UPDATE_DATABASENAME_TAG, DatabaseName);

  // Hopefully the record list was created using ES_TS_NOTTOKEEP_TAG
  // as the record type.

  Records->AddBER(*c_browser);
}


GDT_BOOLEAN 
ZESUPDATEREQUEST::GetAction(INT4 *Action)
{

  if( !c_root->GetSubdirectory(ES_TASKSPECIFIC_TAG, c_browser) )
    return GDT_FALSE;

  ASN1TAGU exttag(ASN1_EXTERNAL);

  if( !c_browser->GetSubdirectory(exttag, c_browser) )
    return GDT_FALSE;

  ASN1TAGU oidtag(ASN1_OBJECTIDENTIFIER);
  STRING OID;

  if( !c_browser->GetOID(oidtag, &OID) || OID != ES_UPDATE_OID )
    return GDT_FALSE;

  if( !c_browser->GetSubdirectory(ES_TS_ESREQUEST_TAG, c_browser) )
    return GDT_FALSE;

  if( !c_browser->GetSubdirectory(ES_TS_TOKEEP_TAG, c_browser) )
    return GDT_FALSE;

  if( !c_browser->GetInt(ES_UPDATE_ACTION_TAG, Action) )
    return GDT_FALSE;

  return GDT_TRUE;
}


GDT_BOOLEAN 
ZESUPDATEREQUEST::GetDatabaseName(STRING *DatabaseName)
{

  if( !c_root->GetSubdirectory(ES_TASKSPECIFIC_TAG, c_browser) )
    return GDT_FALSE;

  ASN1TAGU exttag(ASN1_EXTERNAL);

  if( !c_browser->GetSubdirectory(exttag, c_browser) )
    return GDT_FALSE;

  ASN1TAGU oidtag(ASN1_OBJECTIDENTIFIER);
  STRING OID;

  if( !c_browser->GetOID(oidtag, &OID) || OID != ES_UPDATE_OID )
    return GDT_FALSE;

  if( !c_browser->GetSubdirectory(ES_TS_ESREQUEST_TAG, c_browser) )
    return GDT_FALSE;

  if( !c_browser->GetSubdirectory(ES_TS_TOKEEP_TAG, c_browser) )
    return GDT_FALSE;

  if( !c_browser->GetChar(ES_UPDATE_DATABASENAME_TAG, DatabaseName) )
    return GDT_FALSE;

  return GDT_TRUE;
}


//***************************************************************************
//
//	ZESUPDATERESPONSE - Constructor for ES Update Response PDU
//
//		RefId - Reference Identifier
//		TaskPackage - Task Package record
//		Action - Action performed
//		DatabaseName - Database updated
//		UpdateStatus - Update Status
//***************************************************************************
ZESUPDATERESPONSE::ZESUPDATERESPONSE(	const STRING& RefId,
					ZESTASKPACKAGE *TaskPackage,
					const INT4 Action,
					const STRING& DatabaseName,
					const INT4 UpdateStatus,
					ZDEFAULTDIAGFORMAT *DiagRecord)
  : ZESRESPONSE(RefId,
		ES_DONE_OPERATIONSTATUS,
		NULL,
		TaskPackage)
{

  if ( !TaskPackage )
    return;

  c_root->GetSubdirectory(ES_TASKPACKAGE_TAG, c_browser);

  ASN1TAGU exttag(ASN1_EXTERNAL);

  c_browser->GetSubdirectory(exttag, c_browser);

  c_browser->GetSubdirectory(ES_TP_TASKSPECIFIC_TAG, c_browser);

  c_browser->GetSubdirectory(exttag, c_browser);

  c_browser->GetSubdirectory(ES_TS_TASKPACKAGE_TAG, c_browser);

  BERBROWSER sub(*c_browser);

  c_browser->AddSubdirectory(ES_TS_TOKEEP_TAG, &sub);

  sub.AddInt(ES_UPDATE_ACTION_TAG, Action);
  sub.AddChar(ES_UPDATE_DATABASENAME_TAG, DatabaseName);

  c_browser->AddSubdirectory(ES_TS_TARGETPART_TAG, &sub);

  sub.AddInt(ES_UPDATE_UPDATESTATUS_TAG, UpdateStatus);

  if ( DiagRecord ) {
    sub.AddSubdirectory(ES_UPDATE_GLOBALDIAGNOSTICS_TAG, c_browser);

    ASN1TAGU seqtag(ASN1_SEQUENCE);

    c_browser->AddSubdirectory(seqtag, &sub);
    DiagRecord->AddBER(sub);
  }
}


GDT_BOOLEAN 
ZESUPDATERESPONSE::GetUpdateStatus(INT4 *UpdateStatus)
{
  if( !c_root->GetSubdirectory(ES_TASKPACKAGE_TAG, c_browser) )
    return GDT_FALSE;

  ASN1TAGU exttag(ASN1_EXTERNAL);

  if( !c_browser->GetSubdirectory(exttag, c_browser) )
    return GDT_FALSE;

  ASN1TAGU oidtag(ASN1_OBJECTIDENTIFIER);
  STRING OID;

  if( !c_browser->GetOID(oidtag, &OID) || OID != EXTENDEDSERVICES_OID )
    return GDT_FALSE;

  STRING PackageType;

  if( !c_browser->GetOID(ES_TP_PACKAGETYPE_TAG, &PackageType) ||
      PackageType != ES_UPDATE_OID )
    return GDT_FALSE;

  if( !c_browser->GetSubdirectory(ES_TP_TASKSPECIFIC_TAG, c_browser) )
    return GDT_FALSE;

  if ( !c_browser->GetSubdirectory(exttag, c_browser) )
    return GDT_FALSE;

  if( !c_browser->GetOID(oidtag, &OID) || OID != PackageType )
    return GDT_FALSE;

  if( !c_browser->GetSubdirectory(ES_TS_TASKPACKAGE_TAG, c_browser) )
    return GDT_FALSE;

  if( !c_browser->GetSubdirectory(ES_TS_TARGETPART_TAG, c_browser) )
    return GDT_FALSE;

  if( !c_browser->GetInt(ES_UPDATE_UPDATESTATUS_TAG, UpdateStatus) )
    return GDT_FALSE;

  return GDT_TRUE;
}


/*
GDT_BOOLEAN 
ZSEARCHREQUEST::AddTerm(BERBROWSER & b, const STRING& Term)
{
  INT IsSet=0;
  STRING ETerm=Term;
  FILE *fpa;
  
  
  if(IsBoolean(ETerm)) {
    STRING Term1, Term2, Operator;
    BERBROWSER sub(b);
    b.AddSubdirectory(RPNRPNOP_TAG, &sub);
    if ( !GetTerm(ETerm, &Term1, &Term2) )
      return GDT_FALSE;
     
    AddTerm(sub, Term1);
    AddTerm(sub, Term2);
    sub.AddSubdirectory(OPERATOR_TAG, &sub);
    GetOperator(Term, &Operator);
    Operator.UpperCase();
    if(Operator == "AND")
      sub.AddNULL(AND_TAG);
    else if(Operator == "OR")
      sub.AddNULL(OR_TAG);
    else if(Operator == "ANDNOT")
      sub.AddNULL(ANDNOT_TAG);
    else if(Operator == "NEAR") {
      // 2 == word Unit
      AddProximity(sub, 5, 2, GDT_FALSE);
    }else if(Operator.Search("PROX")){

      CHR VAL[256];
      INT distance;
      CHR cc;
      INT rtype=0;
      GDT_BOOLEAN order=GDT_TRUE;
      CHR ss[10];
      Operator.GetCString(VAL,256);
      size_t kk;
      for(kk=0; kk<strlen(VAL); kk++)
	if(VAL[kk]=='-')
	  VAL[kk]=' ';
      sscanf(VAL,"%s %d %c",ss,&distance,&cc);
      switch(cc){
      case 'W':
	rtype=2;			// Word
	order=GDT_TRUE;			// ordering matters;
	break;
      case 'A':
	rtype=2;			// paragraph
	order=GDT_FALSE;		// order doesn't matter
	break;
      case 'L':
	rtype=5;			// section
	order=GDT_FALSE;		// order doesn't matter
	break;
      case 'P':
	rtype=4;			// paragraph
	order=GDT_FALSE;
	break;
      }
      AddProximity(sub, distance, rtype, order);
    }
  } else {
    BERBROWSER sub(b);
    b.AddSubdirectory(OP_TAG, &sub);
    if(ETerm.Search('#')){
      // handle L and S numbers
      STRING UTerm;
      UTerm=ETerm;
      STRINGINDEX a,c;
      a=UTerm.Search('#');
      c=UTerm.Search('[');
      c--;
      a++;
      UTerm.EraseBefore(a);
      UTerm.EraseAfter(c-1);
      IsSet=1;
      UTerm.UpperCase();
      if(UTerm.Search('L')){
	sub.AddChar(RESULTSETID_TAG, UTerm);
	return GDT_TRUE;
      }
    }
    sub.AddSubdirectory(ATTRIBUTESPLUSTERM_TAG, &sub);
    BERBROWSER sub2(sub);
    BERBROWSER sub3(sub);
    sub.AddSubdirectory(ATTRIBUTELIST_TAG, &sub2);
    INT 	i,attrcount;
    STRLIST AttrList;
    STRING TermOnly;
    if ( !ParseList(ETerm, &AttrList, &TermOnly) )
      return GDT_FALSE;
    attrcount = AttrList.GetTotalEntries();
    INT4 Type, Value;
    STRING Temp;
    for(i=1;i <= attrcount;i+=2) {
      AttrList.GetEntry(i, &Temp);
      Type = Temp.GetInt();
      AttrList.GetEntry(i+1, &Temp);
      Value = Temp.GetInt();
      sub2.AddSubdirectory(ASN1TAGU(ASN1_SEQUENCE), &sub3);
      sub3.AddInt(ATTRIBUTETYPE_TAG, Type);
      sub3.AddInt(NUMERIC_TAG, Value);
    }
    sub.AddChar(TERMGENERAL_TAG, TermOnly);
  }
  return GDT_TRUE;
}


GDT_BOOLEAN 
IsBoolean(const STRING& Term)
{
  STRING Temp;
  Temp = Term;
  Temp.UpperCase();
  if((Temp.Search("AND(")) ||
     (Temp.Search("OR(")) ||
     (Temp.Search("NEAR(")) ||
     (Temp.Search("PROX-")) ||
     (Temp.Search("ANDNOT(")))
    return GDT_TRUE;
  return GDT_FALSE;
}
*/

// Replacement from John Wehle
GDT_BOOLEAN
ZSEARCHREQUEST::AddTerm(BERBROWSER &b, const STRING &Term)
{
  const char *p;
  
  p = Term;
  
  /*
   * Skip leading spaces.
   */
 
  const char *term;
 
  while ( isspace(*p) )
    p++;
 
  term = p;
 
  /*
   * Parse and handle the original KWAQS boolean operator.
   */
 
  STRING Operator;
  STRING Term1;
  STRING Term2;
 
  if ( IsBoolean(term) ) {
    if ( !GetOperator(term, Operator) )
      return GDT_FALSE;
    if ( !GetTerm(term, Term1, Term2) )
      return GDT_FALSE;
    if ( !AddBooleanTerm(b, Operator, Term1, Term2) )
      return GDT_FALSE;
    return GDT_TRUE;
  }
 
  /*
   * Parse infix boolean operators.
   */
 
  GDT_BOOLEAN strip_parentheses;
  const char *q;
  const char *op;
  const char *word;
  int nparentheses;
  int op_nparentheses;
 
  op = (char *)NULL;
  word = (char *)NULL;
  nparentheses = 0;
  op_nparentheses = 0;
  strip_parentheses = (*term == '(') ? GDT_TRUE : GDT_FALSE;
 
  while ( *p ) {
 
    /*
     * Look for an operator outside of parentheses.
     */
 
    if ( isspace(*p) || *p == '(')
      if ( word ) {
	if ( !op || nparentheses <= op_nparentheses)
	  if (StrNCaseCmp(word, "AND", 3) == 0 ||
	      StrNCaseCmp(word, "OR", 2) == 0) {
	    op = word;
	    op_nparentheses = nparentheses;
 
	    /*
	     * Skip the spaces leading up to the next token
	     * forcing the token to be handled as not an
	     * operator.
	     */
 
	    while ( isspace(*p) )
	      p++;
 
	    if (StrNCaseCmp(op, "AND", 3) == 0 &&
		StrNCaseCmp(p, "NOT", 3) == 0 &&
		( isspace(*(p + 3)) || *(p + 3) == '(' || !*(p + 3) )) {
	      p += 3;
	      while ( isspace(*p) )
		p++;
	    }
 
	    if (*p == ')' || !*p ) {
	      cerr << "AddTerm encountered an error parsing ";
	      cerr << '"' << Term << '"' << endl;
	      return GDT_FALSE;
	    }
	  }
	word = (char *)NULL;
      }
 
    if ( isspace(*p) || *p == ')') {
      word = (char *)NULL;
      if ( isalpha(*(p + 1)) )
	word = p + 1;
    }
 
    /*
     * Handle quoting and keep track of parentheses.
     */
 
    if ( !nparentheses && !isspace(*p) && p != term)
      strip_parentheses = GDT_FALSE;
 
    if (*p == '"') {
      if ( (q = strchr(p + 1, '"')) )
	p = q;
    }
    else if (*p == '(') {
      nparentheses++;
    }
    else if (*p == ')') {
      if ( !nparentheses ) {
	cerr << "AddTerm encountered an error parsing ";
	cerr << '"' << Term << '"' << endl;
	return GDT_FALSE;
      }
      --nparentheses;
    }
 
    p++;
  }
 
  /*
   * Skip over the outer parentheses and do a sanity check.
   */
 
  if ( strip_parentheses ) {
    term++;
    while (p != term && isspace(*(p - 1)))
      --p;
    if (p == term || *(p - 1) != ')') {
      cerr << "AddTerm encountered an error parsing ";
      cerr << '"' << Term << '"' << endl;
      return GDT_FALSE;
    }
    --p;
  }
 
  /*
   * Handle infix boolean operator.
   */
 
  if ( op ) {
    for (q = op; isalpha(*q); q++)
      ;
 
    Operator = STRING(op, q - op);
 
    while ( isspace(*q) )
      q++;
 
    if (StrNCaseCmp(op, "AND", 3) == 0 && StrNCaseCmp(q, "NOT", 3) == 0 &&
	( isspace(*(q + 3)) || *(q + 3) == '(' || !*(q + 3) )) {
      Operator = "ANDNOT";
      q += 3;
      while ( isspace(*q) )
	q++;
    }
 
    Term1 = STRING(term, op - term);
    Term2 = STRING(q, p - q);
    if ( !AddBooleanTerm(b, Operator, Term1, Term2) )
      return GDT_FALSE;
    return GDT_TRUE;
  }
 
  /*
   * Handle term.
   */
 
  STRLIST AttrList;
  STRING TermOnly;
 
  if ( !ParseList(STRING(term, p - term), &AttrList, &TermOnly) )
    return GDT_FALSE;
 
  BERBROWSER sub(b);
 
  b.AddSubdirectory(OP_TAG, &sub);
  sub.AddSubdirectory(ATTRIBUTESPLUSTERM_TAG, &sub);
 
  BERBROWSER sub2(sub);
  BERBROWSER sub3(sub);
 
  sub.AddSubdirectory(ATTRIBUTELIST_TAG, &sub2);
 
  INT attrcount;
  INT i;
  INT4 Type;
  INT4 Value;
  STRING Temp;
 
  attrcount = AttrList.GetTotalEntries();
 
  for (i = 1;i <= attrcount; i += 2) {
    AttrList.GetEntry(i, &Temp);
    Type = Temp.GetInt();
    AttrList.GetEntry(i + 1, &Temp);
    Value = Temp.GetInt();
    sub2.AddSubdirectory(ASN1TAGU(ASN1_SEQUENCE), &sub3);
    sub3.AddInt(ATTRIBUTETYPE_TAG, Type);
    sub3.AddInt(NUMERIC_TAG, Value);
  }
 
  sub.AddChar(TERMGENERAL_TAG, TermOnly);
 
  return GDT_TRUE;
}
 
 
static 
GDT_BOOLEAN
IsBoolean(const STRING& Term)
{
 
  if (StrNCaseCmp(Term, "AND(", 4) == 0 ||
      StrNCaseCmp(Term, "OR(", 3) == 0 ||
      StrNCaseCmp(Term, "NEAR(", 5) == 0 ||
      StrNCaseCmp(Term, "PROX-", 5) == 0 ||
      StrNCaseCmp(Term, "ANDNOT(", 7) == 0)
    return GDT_TRUE;
 
  return GDT_FALSE;
}
  

//
// Pass a boolean term of the form:
//
//	and(dog[1=2],cat[1=2])
//
//	GetTerm(1, "and(dog[1=21],cat)", &Term1, &Term2)
//
// On return, Term1="dog[1=21]", Term2="cat"
//
// Replacement from John Wehle
static
GDT_BOOLEAN
GetTerm(const STRING &Term, STRING &Term1, STRING &Term2)
{
  char *p;
  char *q;
  char *question;
  char *term1;
  char *term2;
  int parencount;
 
  Term1 = "";
  Term2 = "";
 
  parencount = 0;
 
  term1 = (char *)NULL;
  term2 = (char *)NULL;
 
  p = question = Term.NewCString();
 
  while ( *p ) {
    switch ( *p ) {
    case '(':
      if (parencount++ == 0) {
	term1 = p + 1;
      }
      break;
 
    case ',':
      if ( !parencount ) {
	cerr << "GetTerm encountered an error parsing ";
	cerr << '"' << Term << '"' << endl;
	delete [] question;
	return GDT_FALSE;
      }
      if (parencount == 1) {
	term2 = p + 1;
	*p = '\0';
      }
      break;
 
    case ')':
      if ( !parencount ) {
	cerr << "GetTerm encountered an error parsing ";
	cerr << '"' << Term << '"' << endl;
	delete [] question;
	return GDT_FALSE;
      }
      if (--parencount == 0) {
	*p = '\0';
	continue;
      }
      break;
 
    case '"':
      if ( (q = strchr(p + 1, '"')) )
	p = q;
      break;
 
    case '[':
      while ( *++p ) {
	if (*p == '"') {
	  if ( (q = strchr(p + 1, '"')) )
	    p = q;
	}
	else if (*p == ']') {
	  p++;
	  break;
	}
      }
      continue;
      /* NOTREACHED */
      break;
 
    default:
      break;
    }
 
    p++;
  }
 
  if ( !term1 || !term2 ) {
    cerr << "GetTerm encountered an error parsing ";
    cerr << '"' << Term << '"' << endl;
    delete [] question;
    return GDT_FALSE;
  }
 
  while ( isspace(*term1) )
    term1++;
 
  while ( isspace(*term2) )
    term2++;
 
  Term1 = term1;
  Term2 = term2;
 
  delete [] question;

  return GDT_TRUE;
}
  

static 
GDT_BOOLEAN
GetOperator(const STRING &Term, STRING &Operator)
{
  STRINGINDEX p;
 
  Operator = Term;
  if ( !(p = Operator.Search("(")) )
    return GDT_FALSE;
 
  Operator.EraseAfter(p - 1);
 
  return GDT_TRUE;
}
 

GDT_BOOLEAN
ZSEARCHREQUEST::AddBooleanTerm(BERBROWSER &b,
			       const STRING &Op, 
			       const STRING &Term1, const STRING &Term2)
{
  BERBROWSER sub(b);
  STRING Operator;

  Operator = Op;
  b.AddSubdirectory(RPNRPNOP_TAG, &sub);
     
  AddTerm(sub, Term1);
  AddTerm(sub, Term2);
 
  sub.AddSubdirectory(OPERATOR_TAG, &sub);
 
  if (Operator ^= "AND")
    sub.AddNULL(AND_TAG);
  else if (Operator ^= "OR")
    sub.AddNULL(OR_TAG);
  else if (Operator ^= "ANDNOT")
    sub.AddNULL(ANDNOT_TAG);
  else if (Operator ^= "NEAR") {
    // 2 == word Unit
    AddProximity(sub, 5, 2, GDT_FALSE);
  }
  else if (Operator.Search("PROX")) {
    CHR VAL[256];
    size_t kk;
 
    Operator.GetCString(VAL, sizeof(VAL));
    for (kk=0; kk < strlen(VAL); kk++)
      if (VAL[kk] == '-')
	VAL[kk]=' ';
 
    /*
     * Skip over PROX.
     */
 
    char c;
    istrstream cparams(VAL, sizeof(VAL));
 
    while ( (cparams >> c) && isalpha(c) )
      ;
 
    /*
     * Parse the parameters.
     */
 
    CHR cc;
    GDT_BOOLEAN order;
    INT distance;
    INT rtype=0;
 
    if ( !(cparams >> distance >> cc) )
      return GDT_FALSE;
 
    switch (cc) {
    case 'W':
      rtype=2;                        // Word
      order=GDT_TRUE;                 // ordering matters;
      break;
 
    case 'A':
      rtype=2;                        // paragraph
      order=GDT_FALSE;                // order doesn't matter
      break;
 
    case 'L':
      rtype=5;                        // section
      order=GDT_FALSE;                // order doesn't matter
      break;
 
    case 'P':
      rtype=4;                        // paragraph
      order=GDT_FALSE;
      break;
 
    default:
      order=GDT_FALSE;
      break;
    }
 
    AddProximity(sub, distance, rtype, order);
  }
 
  return GDT_TRUE;
}
  
 
/* 
void 
GetTerm(const STRING &Term, STRING *Term1, STRING *Term2)
{
  STRING Temp;
  STRINGINDEX p;
  INT inBracket=0;
  Temp = Term;
  p = Temp.Search("(");
  Temp.EraseBefore(p+1);
  p = Temp.SearchReverse(")");
  Temp.EraseAfter(p-1);
  
  CHR *buf, *term1, *term2;
  buf = Temp.NewCString();
  term1 = Temp.NewCString();
  term2 = Temp.NewCString();
  INT 	len=Temp.GetLength(),
  term1pos=0,
  i;
  enum { NONE, INPARENS } State;
  State = NONE;
  GDT_BOOLEAN Done = GDT_FALSE;
  INT parencount=0;
  for(i=0;i < len;i++) {
    switch(State) {
    case NONE:
      switch(buf[i]) {
	
      case '\"':
	break;
      case '[':
	inBracket=1;
	term1[term1pos] = buf[i];
	term1pos++;
	break;
      case ']':
	inBracket=0;
	term1[term1pos] = buf[i];
	term1pos++;
	break;
	
      case '(':
	State = INPARENS;
	parencount++;
	term1[term1pos] = buf[i];
	term1pos++;
	break;
      case ',':
	if(parencount==0 && !inBracket){
	  term1[term1pos] = '\0';
	  strcpy(term2, buf+i+1);
	  Done = GDT_TRUE;
	}else{
	  term1[term1pos] = buf[i];
	  term1pos++;
	  
	}
	break;
      default:
	term1[term1pos] = buf[i];
	term1pos++;
	break;
      }
      break;
    case INPARENS:
      switch(buf[i]) {
      case '\"':
	break;
      case '[':
	inBracket=1;
	break;
      case ']':
	inBracket=0;
	break;
      case '(':
	parencount++;
	break;
      case ')':
	parencount--;
	if(parencount==0)
	  State = NONE;
	break;
      }
      
      term1[term1pos] = buf[i];
      term1pos++;
      break;
    }
    if(Done)
      break;
  }
  *Term1 = term1;
  *Term2 = term2;
}


void 
GetOperator(const STRING &Term, STRING *TheOperator)
{
  STRING Temp;
  Temp = Term;
  STRINGINDEX p = Temp.Search("(");
  Temp.EraseAfter(p-1);
  *TheOperator = Temp;
}
*/


GDT_BOOLEAN
ParseList(const STRING &Term, STRLIST *AttrList, STRING *TermOnly)
{
  char *attrib_list;
  char *p;
  char *q;
  char *term;

  AttrList->Clear();
  *TermOnly = "";

  p = term = Term.NewCString();

  /*
   * Locate the attribute list.
   */

  attrib_list = NULL;

  while ( *p ) {
    if (*p == '"') {
      if ( (q = strchr(p + 1, '"')) )
        p = q;
    }
    else if (*p == '[') {
      attrib_list = p + 1;

      if ( !*attrib_list ) {
        cerr << "ParseList encountered an error parsing ";
        cerr << '"' << Term << '"' << endl;
        delete [] term;
        return GDT_FALSE;
      }

      break;
    }
    p++;
  }

  /*
   * Trim trailing spaces.
   */

  while (p != term && isspace(*(p - 1)) )
    --p;

  /*
   * Check for right truncation.
   */

  if (p != term && *(p - 1) == '*') {
    AttrList->AddEntry("5");
    AttrList->AddEntry("1");
    --p;
  }

  *p = '\0';

  /*
   * Process the attribute list.
   */

  char *type;
  char *value;

  p = attrib_list;

  while (p && *p) {

    /*
     * Attribute type.
     */

    while ( isspace(*p) )
      p++;

    if (*p == ']')
      break;

    type = p;

    while ( isdigit(*p) )
      p++;

    while ( isspace(*p) )
      p++;

    if ((*p != '=' && *p != ',') || p == type) {
      cerr << "ParseList encountered an error parsing ";
      cerr << '"' << Term << '"' << endl;
      delete [] term;
      return GDT_FALSE;
    }

    *p++ = '\0';

    AttrList->AddEntry(type);

    /*
     * Attribute value.
     */

    while ( isspace(*p) )
      p++;

    value = p;

    while ( *p ) {
      if (*p == '"') {
        if ( (q = strchr(p + 1, '"')) )
          p = q;
      }
      else if (*p == ',' || *p == ']') {
        break;
      }
      p++;
    }

    if ((*p != ',' && *p != ']') || p == value) {
      cerr << "ParseList encountered an error parsing ";
      cerr << '"' << Term << '"' << endl;
      delete [] term;
      return GDT_FALSE;
    }

    if (*p == ',')
      *p++ = '\0';
    else
      *p = '\0';

    AttrList->AddEntry(value);
  }

  *TermOnly = term;

  delete [] term;

  return GDT_TRUE;
}

void 
ZSEARCHREQUEST::AddProximity(BERBROWSER &Sub, 
			     INT4 Distance,
			     INT Unit, 
			     GDT_BOOLEAN Order)
{
  ASN1TAG CTag(3);
  BERBROWSER ssub(Sub);
  
  Sub.AddSubdirectory(CTag, &ssub);
  
  CTag = 2;
  ssub.AddInt(CTag, Distance);
  
  CTag.Number = 3;
  ssub.AddInt(CTag, (INT4)Order);
  
  CTag.Number = 4;
  ssub.AddInt(CTag, 2);
  
  CTag.Number = 5;
  ssub.AddSubdirectory(CTag, &ssub);
  
  CTag.Number = 1;
  ssub.AddInt(CTag, Unit);
}


//***************************************************************************
//
//	ZSORTREQUEST - Constructor for Sort Request PDU
//
//		RefId - Reference Identifier
//
//***************************************************************************
ZSORTREQUEST::ZSORTREQUEST(	const STRING& RefId,
				const CHR *InputResultSetNames,
				const STRING& SortedResultSetName,
				const CHR *SortSequence)
  : ZSORTBASE(SORTREQUEST_TAG)
{
  c_root->AddChar(INPUTRESULTSETNAMES_TAG, InputResultSetNames);
  AddSortedResultSetName(SortedResultSetName);
  AddSortSequence();
  INT4 SortRelation=0, CaseSensitive=0;
  if(c_root->GetSubdirectory(SORTSEQUENCE_TAG, c_browser)){
    c_browser->AddInt(SORTRELATION_TAG, SortRelation);
    c_browser->AddInt(CASESENSITIVITY_TAG, CaseSensitive);
    BERBROWSER sub(this);
    STRING Abort, NullValue, Value, SortField;
    c_browser->AddSubdirectory(MISSINGVALUEACTION_TAG, &sub);
    sub.AddChar(ABORT_TAG, Abort);
    sub.AddChar(NULL_TAG, NullValue);
    sub.AddChar(MISSINGVALUEDATA_TAG, Value);
    c_browser->AddSubdirectory(GENERICSORTELEMENT_TAG, c_browser);
    c_browser->AddChar(SORTFIELD_TAG, SortField);
  }
}


//***************************************************************************
//
//	ZSORTRESPONSE - Constructor for Sort Response PDU
//	
//***************************************************************************

ZSORTRESPONSE::ZSORTRESPONSE(	const STRING& RefId,
				const INT4 SortStatus,
				const INT4 SortResponse,
				const CHR *Diagnostics)
  : ZSORTBASE(SORTRESPONSE_TAG)
{	
  AddSortStatus(SortStatus);
  AddSortResponse(SortResponse);
  // AddDiagnostics(Diagnostics);
}
