/* $Id: zpdu.hxx,v 1.4 1997/09/28 16:28:30 cnidr Exp $ */
/************************************************************************
Copyright Notice

Copyright (c) MCNC, Clearinghouse for Networked Information Discovery and
Retrieval, 1994. 

Permission to use, copy, modify, distribute, and sell this software and
its documentation, in whole or in part, for any purpose is hereby granted
without fee, provided that

1. The above copyright notice and this permission notice appear in all
copies of the software and related documentation. Notices of copyright
and/or attribution which appear at the beginning of any file included in
this distribution must remain intact. 

2. Users of this software agree to make their best efforts (a) to return
to MCNC any improvements or extensions that they make, so that these may
be included in future releases; and (b) to inform MCNC/CNIDR of noteworthy
uses of this software. 

3. The names of MCNC and Clearinghouse for Networked Information Discovery
and Retrieval may not be used in any advertising or publicity relating to
the software without the specific, prior written permission of MCNC/CNIDR. 

THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY WARRANTY
OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE. 

IN NO EVENT SHALL MCNC/CNIDR BE LIABLE FOR ANY SPECIAL, INCIDENTAL,
INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES WHATSOEVER
RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT ADVISED OF THE
POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY, ARISING OUT OF OR
IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE. 
************************************************************************/

/*@@@
File:		zpdu.hxx
Version:	1.00
Description:	Z39.50 PDU classes
Author:		Kevin Gamiel, kevin.gamiel@cnidr.org
@@@*/

#ifndef _ZPDU_HXX_
#define _ZPDU_HXX_

#include "ber.hxx"
#include "ztags.hxx"
#include "strlist.hxx"

//***************************************************************************
//
//	class ZPDU - Base class for Z39.50 PDUs
//	
//***************************************************************************

/// Base class for all Z39.50 PDU's

/*@Doc:
\begin{description}

\item[Function]This class provides basic, low-level support for all Z39.50 PDU's.
From this class, we derive base classes for all Z39.50 operations, and from 
those base classes, we derive classes for Z39.50 Request and Response PDU's
\item[Algorithms] Varies by method.

\item[Flow] This is a fundamental c++ class.    Flow is described as needed within the individual method descriptions.
\item[Packaging] This fundamental class is linked and archived as determined by the compiling software.
\item[Security] This class has no security role or function, although individual methods may impact security processes.
\end{description}
*/

class ZPDU : public BERTREE {
 protected:
  BERBROWSER *c_root,	*c_browser;	 

/// Basic Initializer	 
  void Init();
/*@Doc:  Performs basic initialization functions*/

 public:
  
  /// Basic Constructor
  ZPDU(PUCHR Buf, INT4 Length);
/*@Doc:  Make a Basic Z39.50 PDU from the  BER-encoded {\tt Buf}, using included data and tagged according to the data.*/
	
  /// Basic Tagged Constructor
  ZPDU(ASN1TAG Tag);
/*@Doc:  Create empty basic Z39.50 PDU with tag {\tt Tag}*/
	
  /// Basic Destructor
  ~ZPDU();
/*@Doc:  Destroy this PDU, calling destructors for specific PDU types */

	
  /// (Client/Server) Check for Reference ID 
  inline INT HasReferenceId   () const	
    { return c_root->HasTag(  ( 2  )  );};
/*@Doc: Check for the presence of a Reference ID in Request or Response PDU's*/

  /// (Client/Server) Get Reference ID
  inline INT GetReferenceId   (  STRING   *argValue) const	
    { return c_root->GetChar  (  ( 2  )  ,argValue);}; 	
/*@Doc:  Get the Reference ID from the Request or Response PDU*/

  /// (Client/Server) Set Reference ID
  inline INT AddReferenceId   (const   STRING   & argValue) const	
    { return c_root->AddChar  (  ( 2  )  ,argValue);};   
/*@Doc: Set the Reference ID in the Request or Response PDU*/
};
typedef ZPDU   *PZPDU;

//***************************************************************************
//
//	class ZINITBASE - Base class for Z39.50 Initialize PDU
//	
//***************************************************************************

/// Main Initializer Class
/*@Doc:
\begin{description}

\item[Function] ZINITBASE is the base class for Init Request and Response PDU's.  It is derived from the ZPDU class, and implements the extra methods needed by the Init process.
\item[Algorithms] Vary by method

\item[Flow] This is a fundamental c++ class.    Flow is described as needed within the individual method descriptions.
\item[Packaging] This fundamental class is linked and archived as determined by the compiling software.
\item[Security] This class performs all initial security exchanges with the Z39.50 server.
\end{description}
*/
class ZINITBASE : public ZPDU {
 public:
	
  /// Create Tagged Initialize Request/Response PDU
    ZINITBASE(ASN1TAG Tag);
  /*@Doc:  Create an Init Request or Response PDU, based on Tag.  Any ZINITBASE PDU can be either a Request or Response -- all methods function for either case.  The tag is for the ``other side'' to determine the PDU type.*/
  
    /// Convert Buffer to Init Request/Response PDU
      ZINITBASE(PUCHR Buf, INT4 Length);
    /*@Doc:  Convert a character buffer back into a Init PDU*/

    /// (Client/Server) Check For Protocol Version
      inline INT HasProtocolVersion   () const	
	{ return c_root->HasTag(  ( 3  )  );}; 	
/*@Doc: Return 1 if PDU has a Protocol Version field*/

    /// (Client/Server) Get Protocol Version
      inline INT GetProtocolVersion   (  STRING   *argValue) const	
	{ return c_root->GetBitString  (  ( 3  )  ,argValue);}; 	
/*@Doc: Get the protocol version from the Init PDU*/

    /// (Client/Server) Add Protocol Version
      inline INT AddProtocolVersion   (const   STRING   & argValue) const
	{ return c_root->AddBitString  (  ( 3  )  ,argValue);};   
/*@Doc: Add Protocol Version to Init PDU*/ 
     
    /// (Client/Server) Check For Z39.50 Options
      inline INT HasOptions   () const	
	{ return c_root->HasTag(  ( 4  )  );}; 	
	/*@Doc: Check for existance of Z39.50 options bitstring.  Returns 1 on existance.*/

    /// (Client/Server) Get  Z39.50 Options
      inline INT GetOptions   (  STRING   *argValue) const	
	{ return c_root->GetBitString  (  ( 4  )  ,argValue);}; 
/*@Doc: Get Z39.50 Options bitstring from Init PDU*/

    /// (Client/Server) Add  Z39.50 Options	
      inline INT AddOptions   (const   STRING   & argValue) const	
	{ return c_root->AddBitString  (  ( 4  )  ,argValue);};   
/*@Doc: Get Z39.50 Options bitstring from Init PDU*/
    
    /// (Client/Server) Check For Preferred Message Size
      inline INT HasPreferredMessageSize   () const	
	{ return c_root->HasTag(  (  5  )  );}; 	
/*@Doc: Check for existance of Z39.50 Init PreferredMessageSize in Init PDU.  Returns 1 on existance.*/

    /// (Client/Server) Get Preferred Message Size
      inline INT GetPreferredMessageSize   (  INT4   *argValue) const	
	{ return c_root->GetInt  (  (  5  )  ,argValue);}; 	
/*@Doc: Return the PreferredMessageSize in bytes from the Init (Request/Response) PDU*/

    /// (Client/Server) Add Preferred Message Size
      inline INT AddPreferredMessageSize   (  INT4   argValue) const	
	{ return c_root->AddInt  (  (  5  )  ,argValue);};  
/*@Doc: Add the PreferredMessageSize in bytes the Init (Request/Response) PDU to the current PDU */
    
    /// (Client/Server) Check For Largest Supported Record Size
      inline INT HasExceptionalRecordSize   () const	
	{ return c_root->HasTag(  (  6  )  );}; 	
/*@Doc: Check for existance of Exceptional Record Size entry*/

    /// (Client/Server) Get Largest Supported Record Size
      inline INT GetExceptionalRecordSize   (  INT4   *argValue) const	
	{ return c_root->GetInt  (  (  6  )  ,argValue);}; 	
/*@Doc: Get the Exceptional Record Size in bytes from the Init (Request/Response) PDU */

    /// (Client/Server) Add Largest Supported Record Size
      inline INT AddExceptionalRecordSize   (  INT4   argValue) const	
	{ return c_root->AddInt  (  (  6  )  ,argValue);};
    /*@Doc: Add the Exceptional Record Size entry in bytes to the Init (Request/Response) PDU*/
    
    /// (Client/Server) Check For Implementation ID
      inline INT HasImplementationId   () const	
	{ return c_root->HasTag(  (  110  )  );}; 	
/*@Doc: Check for the Implementation ID.  Returns 1 on existance.*/

    /// (Client/Server) Get  Implementation ID
      inline INT GetImplementationId   (  STRING   *argValue) const	
	{ return c_root->GetChar  (  (  110  )  ,argValue);}; 	
/*@Doc: Get the Implementation ID from the Init (Request/Response) PDU */
 	
    /// (Client/Server) Add Implementation ID
      inline INT AddImplementationId   (const   STRING   & argValue) const
	{ return c_root->AddChar  (  (  110  )  ,argValue);};   
/*@Doc: Add the Implementation ID to the Init (Request/Response) PDU*/
    
    /// (Client/Server) Check For Implementation Name
      inline INT HasImplementationName   () const	
	{ return c_root->HasTag(  (  111  )  );}; 	
/*@Doc: Check for existance of Implementation Name. Returns 1 on existance.*/

    /// (Client/Server) Get Implementation Name
      inline INT GetImplementationName   (  STRING   *argValue) const	
	{ return c_root->GetChar  (  (  111  )  ,argValue);}; 
/*@Doc: Get the Implementation Name from the Init (Request/Response) PDU Returns 1 on success.*/

    /// (Client/Server) Add Implementation Name
      inline INT AddImplementationName   (const   STRING   & argValue) const
	{ return c_root->AddChar  (  (  111  )  ,argValue);};   
/*@Doc: Add the Implementation Name to the Init (Request/Response) PDU. Returns 1 on success.*/
        
    /// (Client/Server) Check For Implementation Version
      inline INT HasImplementationVersion   () const	
	{ return c_root->HasTag(  (  112  )  );}; 	
/*@Doc: Check for the existance of the Implementation Version field. Returns 1 on existance.*/

    /// (Client/Server) Get Implementation Version
      inline INT GetImplementationVersion   (  STRING   *argValue) const
	{ return c_root->GetChar  (  (  112  )  ,argValue);}; 	
/*@Doc: Get the Implementation Version from the Init (Request/Response) PDU Returns 1 on success.*/

    /// (Client/Server) Set Implementation Version
      inline INT AddImplementationVersion   (const   STRING   & argValue) const
	{ return c_root->AddChar  (  (  112  )  ,argValue);};   
/*@Doc: Add the Implementation Version to  the Init (Request/Response) PDU. Returns 1 on success */

  };

//***************************************************************************
//
//	class ZINITREQUEST - Z39.50 Initialize Request PDU
//
//***************************************************************************

#include "zrecords.hxx"


/// Init Request Base Class
/*@Doc:
\begin{description}

\item[Function]  This class constructs an Initialize Request from a ZINITBASE object.  The correct tags and request-level methods are implemented here.  This class can decode an Init Request received over a network or in a buffer.

\item[Algorithms]  A simple insertion/retrieval of tags and data into a tree are performed.

\item[Flow] This is a fundamental c++ class.    Flow is described as needed within the individual method descriptions.
\item[Packaging] This fundamental class is linked and archived as determined by the comiling software.
\item[Security] This class performs basic security management with the server.
\end{description}
*/
class ZINITREQUEST : public ZINITBASE {
 private:
 public:
  /// Basic Buffer Constructor
  ZINITREQUEST(PUCHR Buf, INT4 Length);
/*@Doc: Construct an Init Request from a buffer of data*/

  /// Basic Parameter List Constructor
  ZINITREQUEST(	const STRING RefId, 
		   const STRING ProtVer,
		   const STRING Options,
		   const INT4 PrefMsgSize,
		   const INT4 ExceptionalRecSize,
		   const STRING GroupId,
		   const STRING UserId,
		   const STRING Password,
		   const STRING ImpId,
		   const STRING ImpName,
		   const STRING ImpVersion);
/*@Doc: Construct an Init Request from a set of parameters*/

    /// Check For Authentication Entry
      inline INT HasAuthentication  () const	
	{ return c_root->HasTag(   7   );}; 	
/*@Doc: Check for existance of Authentication entry.  Returns 1 if it exists*/

    /// Add Authentication Entry
      inline INT AddAuthentication () const	
	{ return c_root->AddSubdirectory(  7  );};	
/*@Doc: Add an Authentication entry to the Init Request PDU.  Returns 1 on success */

    /// Get Authentication Entry
      inline INT GetAuthentication (BERBROWSER* & Dir) const	
	{ return c_root->GetSubdirectory(  7  ,Dir);}; 
/*@Doc: Get the complete Authentication BERBROWSER entry from the Init Request PDU. Returns 1 on success*/

    /// Get the actual authentication string
      GDT_BOOLEAN GetAuthentication (STRING *Pass);
    /*@Doc: Get the actual Authentication string (in Pass) from the the Init Request PDU.  Note that this simply extracts the string from the Authentication BERBROWSER*/

    /// Get the User Group ID
      GDT_BOOLEAN GetGroupId(STRING *GroupId);
/*@Doc: Return a string containing the Group ID from the Init Request PDU.  Returns GDT\_TRUE on success*/

    /// Get the User login ID
      GDT_BOOLEAN GetUserId(STRING *UserId);
/*@Doc: Fills UserId string with User ID from Init Request PDU.  Returns GDT\_TRUE on success*/

    /// Get the user password
      GDT_BOOLEAN GetPassword(STRING *Password);
/*@Doc: Fills Password string with Password from Init Request PDU.  Returns GDT\_TRUE on success*/
  };




//***************************************************************************
//
//	class ZINITRESPONSE - Z39.50 Initialize Response PDU
//
//***************************************************************************


/// Init Response PDU Class
/*@Doc:
\begin{description}

\item[Function]  This object implements a Z39.50 Init Response PDU.  This class can decode a response PDU received over a network, or can generate one for transmission to the client.
\item[Algorithms] Simple tagged data insertion/retrieval into a tree

\item[Flow] This is a fundamental c++ class.    Flow is described as needed within the individual method descriptions.
\item[Packaging] This fundamental class is linked and archived as determined by the comiling software.
\item[Security] This class has no security role or function.
\end{description}
*/
class ZINITRESPONSE : public ZINITBASE {
public:
///Basic Buffer Constructor
  ZINITRESPONSE(PUCHR Buf, INT4 Length);
/*@Doc: Decode a buffer of data from a file or network into a Present Response*/

///Basic Parameter List Constructor 
  ZINITRESPONSE(	const STRING& RefId, 
		const STRING& ProtVer,
		const STRING& Options,
		const INT4 PrefMsgSize,
		const INT4 ExceptionalRecSize,
		const INT4 Result,
		const STRING& ImpId,
		const STRING& ImpName,
		const STRING& ImpVersion);
/*@Doc: Construct an Init Response from a set of parameters*/
  
  /// Check for Init Status Result
    inline INT HasResult   () const	
      { return c_root->HasTag(  (  12  )  );}; 	
/*@Doc: Checks for the existance of the Z3950 status of the Init operation.  Returns 1 on existance.*/
	
  /// Get the Init Status Result
    inline INT GetResult   (  INT4   *argValue) const	
      { return c_root->GetInt  (  (  12  )  ,argValue);}; 
/*@Doc:  Loads {\tt argValue} with  the Z3950 status of the Init operation.  See the Z39.50 specification for details. Returns 1 on success.*/

  /// Add the Init Status Result
    inline INT AddResult   (  INT4   argValue) const	
      { return c_root->AddInt  (  (  12  )  ,argValue);};   
/*@Doc:  Adds {\tt argValue} as the Init operation status.  Returns 1 on success*/  
  
};



//***************************************************************************
//
//	class ZSEARCHBASE - Base class for Z39.50 Search PDUs
//	
//***************************************************************************


/// Base Class for All Z39.50 Search PDU's
/*@Doc:
\begin{description}

\item[Function] This class provides the base methods for Search Request and
  Response PDU's.  The SearchRequest and SearchResponse classes are derived from
  this base class 
\item[Algorithms] Simple tagged data insertion/retrieval into a tree.

\item[Flow] This is a fundamental c++ class.    Flow is described as needed within the individual method descriptions.
\item[Packaging] This fundamental class is linked and archived as determined by the comiling software.
\item[Security] This class has no security role or function.
\end{description}
*/
class ZSEARCHBASE : public ZPDU {
public:
  
  /// Basic Tagged Constructor
  ZSEARCHBASE(ASN1TAG Tag) : ZPDU(Tag) {}
  /*@Doc: Construct a basic Search Request/Response PDU*/

  /// Basic Buffer Constructor 
  ZSEARCHBASE(PUCHR Buf, INT4 Length) : ZPDU(Buf,Length) {}
  /*@Doc:  Construct a Basic Search Request/Response PDU from a buffer 
    of BER encoded data.*/
};



//***************************************************************************
//
//	class ZSEARCHREQUEST - Z39.50 Search Request PDU
//	
//***************************************************************************

/// Z39.50 Search Request PDU Class
/*@Doc:
\begin{description}

\item[Function] Construct a Search Request PDU from a ZSEARCHBASE object.
\item[Algorithms]

\item[Flow] This is a fundamental c++ class.    Flow is described as needed within the individual method descriptions.
\item[Packaging] This fundamental class is linked and archived as determined by the comiling software.
\item[Security] This class has no security role or function.
\end{description}
*/
class ZSEARCHREQUEST : public ZSEARCHBASE {
public:
///Basic Buffer Constructor
  ZSEARCHREQUEST(PUCHR Buf, INT4 Length) : ZSEARCHBASE(Buf,Length) {}
/*@Doc: Create a Search Request from a buffer of BER encoded data.  Typically used by servers to create a Search Request object for decoding from network data*/

///Basic Parameter List Constructor
  ZSEARCHREQUEST(	const STRING& RefId,
		 const INT4 SmallSetUpperBound,
		 const INT4 LargeSetLowerBound,
		 const INT4 MedSetPresentNum,
		 const INT4 ReplaceIndicator,
		 const STRING& ResultSetName,
		 const STRING& DatabaseNames,
		 const CHR *SmallElementSetNames,
		 const CHR *MediumElementSetNames,
		 const STRING& PrefRecSyntax,
		 const STRING& Query,
		 const STRING& AttrSetId,
		 const STRING& u,
		 const INT QueryType=101);
/*@Doc:  Construct Search Request from parameter list*/  

///Check for Small Set Upper Bound
  inline INT HasSmallSetUpperBound   () const	
    { return c_root->HasTag(  (  13  )  );}; 	
 /*@Doc: Check for existance of SmallSetUpperBound Entry.  
    SmallSetUpperBound defines the Maximum number of results that may be in a 
    result set before the system considers it to {\bf not} be a small-set.*/

  /// Get SmallSetUpperBound Entry
    inline INT GetSmallSetUpperBound   (  INT4   *argValue) const	
      { return c_root->GetInt  (  (  13  )  ,argValue);}; 	
/*@Doc:  Get the SmallSetUpperBound entry.  Returns 1 on success, 0 on failure*/

  /// Add SmallSetUpperBound Entry
    inline INT AddSmallSetUpperBound   (  INT4   argValue) const	
      { return c_root->AddInt  (  (  13  )  ,argValue);};  
/*@Doc:  Add a Small Set Upper Bound to the Search Request PDU.  This sets the maximum size of a ``small'' result set*/ 
  
///Check for Large Set Lower Bound  
  inline INT HasLargeSetLowerBound   () const	
    { return c_root->HasTag(  (  14  )  );}; 
/*@Doc: Check for existance of LargeSetLowerBound Entry.  If a result 
    set contains {\bf more} records than specified in LargeSetLowerBound, it is 
    considered to be a large-set.*/

  /// Get LargeSetLowerBound Entry
    inline INT GetLargeSetLowerBound   (  INT4   *argValue) const	
      { return c_root->GetInt  (  (  14  )  ,argValue);}; 	
/*@Doc:  Get the LargeSetLowerBound entry.  Returns 1 on success, 0 on failure*/

  /// Add LargeSetLowerBound Entry
    inline INT AddLargeSetLowerBound   (  INT4   argValue) const	
      { return c_root->AddInt  (  (  14  )  ,argValue);}; 
/*@Doc:  Add a Large Set Lower Bound entry.  This sets the minimum size of a ``large'' result set.*/
   
  /// Check for Medium Set Present Number
  inline INT HasMediumSetPresentNumber   () const	
    { return c_root->HasTag(  (  15  )  );}; 	
/*@Doc: Check for existance of MediumSetPresentNumber Entry.  If the 
    count of entries in a result set is greater than SmallSetLowerBound and less 
    than LargeSetUpperBound, it is a medium-set.  MediumSetPresentNumber 
    specifies the number of records to return if the result set is a medium-set.*/

  /// Get MediumSetPresentNumber Entry
  inline INT GetMediumSetPresentNumber   (  INT4   *argValue) const	
      { return c_root->GetInt  (  (  15  )  ,argValue);}; 
/*@Doc: Load argValue with the Medium SetPresent  Number.  Returns 1 on success.*/
 
  /// Add MediumSetPresentNumber Entry
  inline INT AddMediumSetPresentNumber   (  INT4   argValue) const	
      { return c_root->AddInt  (  (  15  )  ,argValue);}; 
/*@Doc: Add argValue as the MediumSetPresent Number.  Returns 1 on success*/
  
 ///Check for Replace Indicator
  inline INT HasReplaceIndicator   () const	
    { return c_root->HasTag(  (  16  )  );}; 	
 /*@Doc: Check for existance of ReplaceIndicator Entry.  The ReplaceIndicator specifies the action to be taken if a result of the name given by
    ResultSetName already exists.  If ReplaceIndicator is ON, the result set
    will be replaced byt he new result set.  if it is OFF, the search is not
    processed, a diagnostic error message is returned, and the existing result set
    is undisturbed. */

  /// Get ReplaceIndicator Entry
  inline INT GetReplaceIndicator   (  INT4   *argValue) const	
      { return c_root->GetInt  (  (  16  )  ,argValue);}; 	
  /// Add ReplaceIndicator Entry
  inline INT AddReplaceIndicator   (  INT4   argValue) const	
      { return c_root->AddInt  (  (  16  )  ,argValue);};   
  
/// Check for Result Set Name
  inline INT HasResultSetName   () const	
    { return c_root->HasTag(  (  17  )  );}; 
/*@Doc: Check for existance of ResultSetName Entry.  ResultSetName specifies a name to be given to the result set (to be created by the query) so that
    it may subsequently be referenced (within the same Z-association).*/

  /// Get ResultSetName Entry
  inline INT GetResultSetName   (  STRING   *argValue) const	
      { return c_root->GetChar  (  (  17  )  ,argValue);}; 
/*@Doc: Get the Result Set Name.  Put it in {\tt argValue} and return 1 on success*/

  /// Add ResultSetName Entry
  inline INT AddResultSetName   (const   STRING   & argValue) const	
      { return c_root->AddChar  (  (  17  )  ,argValue);};   
/*@Doc:  Add the Result Set Name in {\tt argvalue} to the PDU.  Return 1 on success*/
    
  /// Check for existence of Database Names Entry
    inline INT HasDatabaseNames  () const	
      { return c_root->HasTag(   18   );}; 	
/*@Doc: Return 1 if there is a Database entry */
	
  inline INT AddDatabaseNames () const	
    { return c_root->AddSubdirectory(  18  );};	
/*@Doc:  Add an empty Database Names BERBROWSER entry. Returns 1 on success.*/

/// Get the Database Names Entry.
  inline INT GetDatabaseNames (BERBROWSER* & Dir) const	
    { return c_root->GetSubdirectory(  18  ,Dir);}; 
/*@Doc: Get the Database Names BERBROWSER.  Returns 1 on success*/
  
  /// Get The Database Names string
  void GetDatabaseName(STRING *DBName); 
/*@Doc: Fill DBName with the actual Database Names string.  Only gets first entry in database names list*/
  
/// Check existance of Small Set Element Name entry
  inline INT HasSmallSetElementSetNames  () const	
    { return c_root->HasTag(   100   );}; 	
/*@Doc: Returns 1 if Small Set Element Name entry exists*/

/// Add the Small Set Element Name entry
  inline INT AddSmallSetElementSetNames () const	
    { return c_root->AddSubdirectory(  100  );};	
/*@Doc: Add an empty Small Set Element Name browser*/

/// Get the Small Set Element Name entry
  inline INT GetSmallSetElementSetNames (PBERBROWSER & Dir) const	
    { return c_root->GetSubdirectory(  100  ,Dir);}; 
/*@Doc: Get the Small Set Element name BERBROWSER*/
  
/// Check existance of Medium Set Element Name entry  
  inline INT HasMediumSetElementSetNames  () const	
    { return c_root->HasTag(   101   );}; 	
/*@Doc: Returns 1 if Medium  Set Element Name entry exists*/

/// Add the Medium Set Element Name entry
  inline INT AddMediumSetElementSetNames () const	
    { return c_root->AddSubdirectory(  101  );};	
/*@Doc: Add an empty Medium Set Element Name browser*/

/// Get the Medium Set Element Name entry
  inline INT GetMediumSetElementSetNames (PBERBROWSER & Dir) const	
    { return c_root->GetSubdirectory(  101  ,Dir);}; 
  /*@Doc: Get the Medium Set Element name BERBROWSER*/
  
/// Check for Preferred Record Syntax
  inline INT HasPreferredRecordSyntax   () const	
    { return c_root->HasTag(  (  104  )  );}; 	
/*@Doc: Check for existenc eof Preferred Record Syntax entry.  Returns 1 on existance*/	

/// Get Preferred Record Syntax
  inline INT GetPreferredRecordSyntax   (  STRING   *argValue) const	
    { return c_root->GetOID  (  (  104  )  ,argValue);}; 	
/*@Doc:  Gets the preferred record syntax for piggybacked records*/
	
/// Add Preferred Record Syntax
  inline INT AddPreferredRecordSyntax   (  STRING   argValue) const	
    { return c_root->AddOID  (  (  104  )  ,argValue);};   
/*@Doc: Adds a preferred record syntax for piggybacked records*/ 
  
/// Check For Query
  inline INT HasQuery  () const	{ 
    return c_root->HasTag(   21   );}; 	
/*@Doc:  Returns 1 if PDU contains a query structure*/

/// Add a query subdirectory to PDU
  inline INT AddQuery () const	{ 
    return c_root->AddSubdirectory(  21  );};	
/*@Doc:  Add a subdirectory tagged to contain a query.  Returns 1 on success, 0 if failure. Use GetSubdirectory(Tag,*dir) to retrieve the empty directory for processing.*/

/// Get a query BERBROWSER from PDU
  inline INT GetQuery (PBERBROWSER & Dir) const	
    { return c_root->GetSubdirectory(  21  ,Dir);}; 
/*@Doc: Load {\tt Dir} with the Query BERBROWSER.  Return 1 on success.*/
  
  /// Add Search Term to Search Request
  GDT_BOOLEAN AddTerm(BERBROWSER & b, const STRING& Term);
/*@Doc:  Adds the {\tt Term}, in KWAQS format, to the Z39.50 RPN Query Structure*/

  /// Set the user id
  void SetuserId(const STRING& id);
/*@Doc:  Adds the user identification string to the request*/
  
  /// Add Term Proximity information to Search Request
  void AddProximity(BERBROWSER &Sub, INT4 Distance, INT Unit, 
		    GDT_BOOLEAN Order);
/*@Doc:  Adds term proximity information to the Z39.50 Search proximity Operator*/

private:
  /// This is required for carrying User ID for Annotations, etc.
  CHR GlobalUserId[256];
 
  GDT_BOOLEAN AddBooleanTerm(BERBROWSER &b,
			     const STRING &Operator, 
			     const STRING &Term1, const STRING &Term2);
};


//***************************************************************************
//
//	class ZSEARCHRESPONSE - Z39.50 Search Response PDU
//	
//***************************************************************************


/// Z39.50 Search Response PDU Class
/*@Doc:
\begin{description}

\item[Function] Manage Z39.50 Search Response PDU's.
\item[Algorithms] Basic tagged data insertion/retrieval into a tree.

\item[Flow] This is a fundamental c++ class.    Flow is described as needed within the individual method descriptions.
\item[Packaging] This fundamental class is linked and archived as determined by the compiling software.
\item[Security] This class has no security role or function.
\end{description}
*/
class ZSEARCHRESPONSE : public ZSEARCHBASE {
public:
///Basic Buffered Constructor
  ZSEARCHRESPONSE(PUCHR Buf, INT4 Length) : ZSEARCHBASE(Buf,Length) {}
/*@Doc:  Construct Search Response from buffer of BER encoded data*/

///Basic Parameter List Constructor
  ZSEARCHRESPONSE(const STRING& RefId,
		  const INT4 ResultCount,
		  const INT4 NumberOfRecordsReturned,
		  const INT4 NextResultSetPosition,
		  const INT4 SearchStatus,
		  const INT4 ResultSetStatus,
		  const INT4 PresentStatus,
		  ZRECORDLIST *Records);
/*@Doc:  Construct Search Response from parameter list*/

///Check for ResultSetCount entry
  inline INT HasResultCount   () const	
    { return c_root->HasTag(  (  23  )  );}; 	
/*@Doc:  Return 1 if Result Set Country exists*/

///Get Result Set Count	
  inline INT GetResultCount   (  INT4   *argValue) const	
    { return c_root->GetInt  (  (  23  )  ,argValue);}; 	
/*@Doc: Put count of entries for current result set in {\tt argValue}.  Return 1 on success.*/

///Add Result Set Count
  inline INT AddResultCount   (  INT4   argValue) const	
    { return c_root->AddInt  (  (  23  )  ,argValue);};   
/*@Doc:  Set the count of entries in the result set.  Return 1 on success*/

  
/// Check for existance of entry
  inline INT HasNumberOfRecordsReturned   () const	
    { return c_root->HasTag(  (  24  )  );}; 	
/*@Doc: Returns 1 if entry exists*/

  inline INT GetNumberOfRecordsReturned   (  INT4   *argValue) const	
    { return c_root->GetInt  (  (  24  )  ,argValue);}; 	
  inline INT AddNumberOfRecordsReturned   (  INT4   argValue) const	
    { return c_root->AddInt  (  (  24  )  ,argValue);};   

  /// Check for existance of entry
  inline INT HasNextResultSetPosition   () const	
    { return c_root->HasTag(  (  25  )  );}; 	
 /*@Doc: Returns 1 if entry exists*/

  inline INT GetNextResultSetPosition   (  INT4   *argValue) const	
    { return c_root->GetInt  (  (  25  )  ,argValue);}; 	
  inline INT AddNextResultSetPosition   (  INT4   argValue) const	
    { return c_root->AddInt  (  (  25  )  ,argValue);};   

/// Check for existance of entry  
  inline INT HasSearchStatus   () const	
    { return c_root->HasTag(  (  22  )  );}; 	
/*@Doc: Returns 1 if entry exists*/

  inline INT GetSearchStatus   (  INT4   *argValue) const	
    { return c_root->GetInt  (  (  22  )  ,argValue);}; 	
  inline INT AddSearchStatus   (  INT4   argValue) const	
    { return c_root->AddInt  (  (  22  )  ,argValue);};   
  
/// Check for existance of entry
  inline INT HasResultSetStatus   () const	
    { return c_root->HasTag(  (  26  )  );}; 	
/*@Doc: Returns 1 if entry exists*/

///Get Result Set Status
  inline INT GetResultSetStatus   (  INT4   *argValue) const	
    { return c_root->GetInt  (  (  26  )  ,argValue);}; 	
/*@Doc: Get Result Set Status into {\tt argValue}.  See the Z39.50 specification for possible values*/
 	
/// Add Result Set Status
  inline INT AddResultSetStatus   (  INT4   argValue) const	
    { return c_root->AddInt  (  (  26  )  ,argValue);};   
/*@Doc: Add a Z39.50 Result Set Status value tot he PDU*/

/// Check for existance of entry   
  inline INT HasPresentStatus   () const	
    { return c_root->HasTag(  (  27  )  );}; 	
/*@Doc: Returns 1 if entry exists*/
 	
///Get the Present Status
  inline INT GetPresentStatus   (  INT4   *argValue) const	
    { return c_root->GetInt  (  (  27  )  ,argValue);}; 	
/*@Doc: Get the Z39.50 Present Status in {\tt argValue}*/

///Add Present Status
  inline INT AddPresentStatus   (  INT4   argValue) const	
    { return c_root->AddInt  (  (  27  )  ,argValue);};   
/*@Doc: Add the Present Status tot he PDU*/
  
  /// Get Search Posting Counts from Search Response. Formatted into HTML.
    void GetCASPostings(STRING *Buffer);
/*@Doc: Return the CAS-specific search postings from Search Response. Formatted into HTML.*/
};




//***************************************************************************
//
//	class ZPRESENTBASE - Base class for Z39.50 Present PDUs
//	
//***************************************************************************

/// Main Z39.50 Present Request/Response Base Class
/*@Doc:
\begin{description}

\item[Function]  Basic Class for Z39.50 Present Operations
\item[Algorithms] Basic tagged data insertion/retrieval into a tree.
\item[Flow] This is a fundamental c++ class.    Flow is described as needed within the individual method descriptions.
\item[Packaging] This fundamental class is linked and archived as determined by the compiling software.
\item[Security] This class has no security role or function.
\end{description}
*/
class ZPRESENTBASE : public ZPDU {
public:
///Basic Tagged Constructor
  ZPRESENTBASE(ASN1TAG Tag) : ZPDU(Tag) {}
/*@Doc: Construct a basic Present Request/Response PDU*/

///Basic Buffer Constructor
  ZPRESENTBASE(PUCHR Buf, INT4 Length) : ZPDU(Buf,Length) {}
/*@Doc:  Construct a Basic Present Request/Response PDU from a buffer of BER encoded data.*/
};


//***************************************************************************
//
//	class ZPRESENTREQUEST - Z39.50 Present Request PDU
//	
//***************************************************************************


/// Z39.50 Present Request PDU Class
/*@Doc:
\begin{description}

\item[Function]  Implements a Z39.50 Present Request
\item[Algorithms] Basic tagged data insertion/retrieval into a tree.

\item[Flow] This is a fundamental c++ class.    Flow is described as needed within the individual method descriptions.
\item[Packaging] This fundamental class is linked and archived as determined by the compiling software.
\item[Security] This class has no security role or function.
\end{description}
*/
class ZPRESENTREQUEST : public ZPRESENTBASE {
public:
///Basic Buffer Constructor
  ZPRESENTREQUEST(PUCHR Buf, INT4 Length) : ZPRESENTBASE(Buf,Length) {}
/*@Doc: Create a Present Request from a buffer of BER encoded data.  Typically used by servers to create a Present Request object for decoding from network data*/

///Basic Parameter List Constructor
  ZPRESENTREQUEST(const STRING& RefId,
		  const STRING& ResultSetId,
		  const INT4 ResultSetStartPoint,
		  const INT4 NumberOfRecordsRequested,
		  const STRING& ElementSetName,
		  const STRING& PreferredRecordSyntax);
/*@Doc: Create a Present request from a set of parameters.  Typically used by clients to create a Present Request object for encoding and transmission to server.*/
  
//Check for Result Set ID entry 
  inline INT HasResultSetId   () const	
    { return c_root->HasTag(  (  31  )  );}; 	
/*@Doc: Return 1 if entry exists*/

///Get entry
  inline INT GetResultSetId   (  STRING   *argValue) const	
    { return c_root->GetChar  (  (  31  )  ,argValue);}; 	
/*@Doc: Return 1 if entry successfully retrieved into {\tt argValue}*/

///Add entry
  inline INT AddResultSetId   (const   STRING   & argValue) const	
    { return c_root->AddChar  (  (  31  )  ,argValue);};   

///Check for Result Set Start Point entry
  inline INT HasResultSetStartPoint   () const	
    { return c_root->HasTag(  (  30  )  );}; 	
/*@Doc: Return 1 if entry exists*/

///Get entry
  inline INT GetResultSetStartPoint   (  INT4   *argValue) const	
    { return c_root->GetInt  (  (  30  )  ,argValue);}; 	
/*@Doc: Return 1 if entry successfully retrieved into {\tt argValue}*/

///Add Entry
  inline INT AddResultSetStartPoint   (  INT4   argValue) const	
    { return c_root->AddInt  (  (  30  )  ,argValue);};   
/*@Doc:  Return 1 if entry successfully added to PDU*/ 
  
///Check for entry
  inline INT HasNumberOfRecordsRequested   () const	
    { return c_root->HasTag(  (  29  )  );}; 	
/*@Doc: Return 1 if entry exists*/

///Retrieve entry
  inline INT GetNumberOfRecordsRequested   (  INT4   *argValue) const	
    { return c_root->GetInt  (  (  29  )  ,argValue);}; 	
/*@Doc: Return 1 if entry successfully retrieved into {\tt argValue}*/

///Add entry	
  inline INT AddNumberOfRecordsRequested   (  INT4   argValue) const	
    { return c_root->AddInt  (  (  29  )  ,argValue);};   
/*@Doc:  Return 1 if entry successfully added to PDU*/ 
 
///Check for existance of entry
  inline INT HasPreferredRecordSyntax   () const	
    { return c_root->HasTag(  (  104  )  );}; 	
/*@Doc: Return 1 if entry exists*/
 	
///Get entry
  inline INT GetPreferredRecordSyntax   (  STRING   *argValue) const	
    { return c_root->GetOID  (  (  104  )  ,argValue);}; 	
/*@Doc: Return 1 if entry successfully retrieved into {\tt argValue}*/

///Add entry
  inline INT AddPreferredRecordSyntax   (  STRING   argValue) const	
    { return c_root->AddOID  (  (  104  )  ,argValue);};   
/*@Doc:  Return 1 if entry successfully added to PDU*/
 
///Get Element Set Name from PDU
  void GetElementSetName(STRING *ElementSetName);

///Check for existance of entry
  inline INT HasAdditionalRanges  () const	
    { return c_root->HasTag(   212  );}; 	
/*@Doc: Return 1 if entry exists*/
	
///Add entry
  inline INT AddAdditionalRanges () const	
    { return c_root->AddSubdirectory(  212 );};	
/*@Doc:  Return 1 if entry successfully added to PDU*/

///Get entry
  inline INT GetAdditionalRanges (PBERBROWSER & Dir) const	
    { return c_root->GetSubdirectory(  212 ,Dir);}; 
/*@Doc: Return 1 if entry successfully retrieved into {\tt Dir}*/ 

///Check for existance of entry
  inline INT HasCompSpec  () const	
    { return c_root->HasTag(   209  );}; 	
/*@Doc: Return 1 if entry exists*/
	
///Add entry
   inline INT AddCompSpec () const	
    { return c_root->AddSubdirectory(  209 );};	
/*@Doc:  Return 1 if entry successfully added to PDU*/

///Get entry
  inline INT GetCompSpec (PBERBROWSER & Dir) const	
    { return c_root->GetSubdirectory(  209 ,Dir);}; 
/*@Doc: Return 1 if entry successfully retrieved into {\tt Dir}*/
};


//***************************************************************************
//
//	class ZPRESENTRESPONSE - Z39.50 Present Response PDU
//	
//***************************************************************************


/// Z39.50 Present Response PDU Base Class
/*@Doc:
\begin{description}

\item[Function] Implements Z39.50 Present Response
\item[Algorithms]

\item[Flow] This is a fundamental c++ class.    Flow is described as needed within the individual method descriptions.
\item[Packaging] This fundamental class is linked and archived as determined by the compiling software.
\item[Security] This class has no security role or function.
\end{description}
*/
class ZPRESENTRESPONSE : public ZPRESENTBASE {
public:
///Basic Buffered Constructor
  ZPRESENTRESPONSE(PUCHR Buf, INT4 Length) : ZPRESENTBASE(Buf,Length) {}
/*@Doc: Construct a Present Response from a buffer of data*/

///Basic Constructor
  ZPRESENTRESPONSE(const STRING& RefId,
		   const INT4 NumberOfRecordsReturned,
		   const INT4 NextResultSetPosition,
		   const INT4 PresentStatus,
		   ZRECORDLIST *Records);
/*@Doc: Construct a Present Response from a set of parameters*/
  
///Check for existance of entry  
  inline INT HasNextResultSetPosition   () const	
    { return c_root->HasTag(  (  25  )  );}; 	
/*@Doc: Return 1 if entry exists*/
	
///Get Entry
  inline INT GetNextResultSetPosition   (  INT4   *argValue) const	
    { return c_root->GetInt  (  (  25  )  ,argValue);}; 	
/*@Doc: Put entry value in {\tt argValue} and return 1 on success*/	

///Add entry
  inline INT AddNextResultSetPosition   (  INT4   argValue) const	
    { return c_root->AddInt  (  (  25  )  ,argValue);};   
/*@Doc: Add entry to PDU*/
  
///Check for existance of entry
  inline INT HasNumberOfRecordsReturned   () const	
    { return c_root->HasTag(  (  24  )  );}; 	
/*@Doc: Return 1 if entry exists*/

///Get Entry	
  inline INT GetNumberOfRecordsReturned   (  INT4   *argValue) const	
    { return c_root->GetInt  (  (  24  )  ,argValue);}; 	
/*@Doc: Put entry value in {\tt argValue} and return 1 on success*/	

 ///Add entry	
  inline INT AddNumberOfRecordsReturned   (  INT4   argValue) const	
    { return c_root->AddInt  (  (  24  )  ,argValue);};   
/*@Doc: Add entry to PDU*/
 
///Check for existance of entry
  inline INT HasPresentStatus   () const	
    { return c_root->HasTag(  (  27  )  );}; 	
/*@Doc: Return 1 if entry exists*/

///Get Entry	
  inline INT GetPresentStatus   (  INT4   *argValue) const	
    { return c_root->GetInt  (  (  27  )  ,argValue);}; 	
/*@Doc: Put entry value in {\tt argValue} and return 1 on success*/	

///Add entry
  inline INT AddPresentStatus   (  INT4   argValue) const	
    { return c_root->AddInt  (  (  27  )  ,argValue);};   
/*@Doc: Add entry to PDU*/

};

//***************************************************************************
//
//	class ZSCANBASE - Base class for Z39.50 Scan PDUs
//	
//***************************************************************************


/// Main Z39.50 SCAN PDU Base Class 
/*@Doc: Scan is used to scan an ordered termlist.  The ordering of
  the termlist is target (server) defined.  The origin (client) specifies a termlist to scan and a starting term, the size of the scanning steps, and the desired number of entries and position of the starting term in the response.  
*/

class ZSCANBASE : public ZPDU {
public:
  /// Request Constructor
  ZSCANBASE(ASN1TAG Tag) : ZPDU(Tag) {}
  /*@Doc: Create Scan PDU from parameter list*/

  /// Response Constructor
  ZSCANBASE(PUCHR Buf, INT4 Length) : ZPDU(Buf,Length) {}
  /*@Doc: Create SCAN PDU from buffer of BER encoded data*/
  
};

//***************************************************************************
//
//	class ZSCANREQUEST - Z39.50 Scan Request PDU
//	
//***************************************************************************


/// Main Z39.50 Scan Request PDU
/*@Doc:
\begin{description}

\item[Function] Generates a Z39.50 Scan Request.  IN GPI, this is the equivalet of a Messenger EXPAND operation.
\item[Algorithms]

\item[Flow] This is a fundamental c++ class.    Flow is described as needed within the individual method descriptions.
\item[Packaging] This fundamental class is linked and archived as determined by the compiling software.
\item[Security] This class has no security role or function.
\end{description}
*/
class ZSCANREQUEST : public ZSCANBASE {
public:
  /// Basic Buffer Constructor
    ZSCANREQUEST(PUCHR Buf, INT4 Length) : ZSCANBASE(Buf,Length) {}
    /// Client constructs a Scan Request
/*@Doc: Create Scan request from buffer of BER encoded data*/

    /// Basic Constructor
      ZSCANREQUEST(   const CHR *RefId,
		   const CHR *DatabaseNames,
		   const CHR *AttributeSet,
		   const CHR *TermList,
		   INT4 StepSize,
		   INT4 NumTermsRequested,
		   INT4 PrefPosInResponse);
/*@Doc: Create Scan request from parameter list*/

    /// Check for existence of database names
      inline INT HasDatabaseNames  () const	
	{ return c_root->HasTag(   3   );}; 	
/*@Doc: Return 1 if entry exists*/
	
    /// Add database names
      inline INT AddDatabaseNames () const	
	{ return c_root->AddSubdirectory(  3  );};	
/*@Doc: Return 1 on success*/

///Get Database Names
    inline INT GetDatabaseNames (PBERBROWSER & Dir) const	
      { return c_root->GetSubdirectory(  3  ,Dir);}; 
/*@Doc: Get DatabaseNames BERBROWSER.  Return 1 on success*/

///Check existance of entry    
    inline INT HasAttributeSet   () const	
      { return c_root->HasTag(  (  6  )  );}; 	
/*@Doc: Return 1 if entry exists*/

///Get Attribute Set
    inline INT GetAttributeSet   (  STRING   *argValue) const	
      { return c_root->GetOID  (  (  6  )  ,argValue);}; 	
/*@Doc: Put entry value in {\tt argValue} and return 1 on success*/	

///Add Attribute Set OID
    inline INT AddAttributeSet   (  STRING   argValue) const	
      { return c_root->AddOID  (  (  6  )  ,argValue);};   
/*@Doc: Adds the attribute set object identifier*/  

///Check existance of entry  
    inline INT HasTermListAndStartPoint  () const	
      { return c_root->HasTag(   102   );}; 	
/*@Doc: Return 1 if entry exists*/

///Add TermList
    inline INT AddTermListAndStartPoint () const	
      { return c_root->AddSubdirectory(  102  );};	
/*@Doc: Add an empty TermList BERBROWSER*/

    inline INT GetTermListAndStartPoint (PBERBROWSER & Dir) const	
      { return c_root->GetSubdirectory(  102  ,Dir);}; 
/*@Doc: Get TermList BERBROWSER and return 1 on success*/	

 ///Check existance of entry  
    inline INT HasStepSize   () const	
      { return c_root->HasTag(  (  5  )  );}; 	
/*@Doc: Return 1 if entry exists*/	

///Get SCAN step size
    inline INT GetStepSize   (  INT4   *argValue) const	
      { return c_root->GetInt  (  (  5  )  ,argValue);}; 	
/*@Doc: Returns 1 on success, and the size of scan steps in {\tt argValue}*/
	
///Add the Scan Step Size
    inline INT AddStepSize   (  INT4   argValue) const	
      { return c_root->AddInt  (  (  5  )  ,argValue);};   
/*@Doc: Set the step size of the SCAN operation*/

  ///Check existance of entry  
    inline INT HasNumberOfTermsRequested   () const	
      { return c_root->HasTag(  (  6  )  );}; 	
/*@Doc: Return 1 if entry exists*/

///Get the number of terms to return
    inline INT GetNumberOfTermsRequested   (  INT4   *argValue) const	
      { return c_root->GetInt  (  (  6  )  ,argValue);}; 	
/*@Doc:  Put th ecount of terms to return from SCAN in {\tt argValue}.*/

///Add Number of Requested Scan Terms
    inline INT AddNumberOfTermsRequested   (  INT4   argValue) const	
      { return c_root->AddInt  (  (  6  )  ,argValue);};   
/*@Doc:  Set the number of scanned terms to return */


///Check existance of entry  
    inline INT HasPreferredPositionInResponse   () const	
      { return c_root->HasTag(  (  7  )  );}; 	
/*@Doc: Return 1 if entry exists*/

///Get the response position to start with
    inline INT GetPreferredPositionInResponse   (  INT4   *argValue) const
      { return c_root->GetInt  (  (  7  )  ,argValue);};
/*@Doc: Return the position in the list of scanned terms to start returning from*/

///Add the Preferred Position
    inline INT AddPreferredPositionInResponse   (  INT4   argValue) const
      { return c_root->AddInt  (  (  7  )  ,argValue);};   
/*@Doc: Set the position at which SCAN starts returning terms*/
  };              



//***************************************************************************
//
//	class ZSCANRESPONSE - Z39.50 Scan Response PDU
//	
//***************************************************************************


/// Z39.50 Scan Response PDU
/*@Doc:
\begin{description}

\item[Function] Generates a Z39.50 Scan Response.  IN GPI, this is the equivalent to a Messenger EXPAND operation.
\item[Algorithms]

\item[Flow] This is a fundamental c++ class.    Flow is described as needed within the individual method descriptions.
\item[Packaging] This fundamental class is linked and archived as determined by the compiling software.
\item[Security] This class has no security role or function.
\end{description}
*/

class ZSCANRESPONSE : public ZSCANBASE {
public:
///Basic Buffer Constructor
  ZSCANRESPONSE(PUCHR Buf, INT4 Length) : ZSCANBASE(Buf,Length) {}
/*@Doc: Construct a Scan Request/Response PDU from a buffer of data*/

///Basic Parameter Constructor
  ZSCANRESPONSE(	const CHR *RefId,
		INT4 StepSize,
		INT4 ScanStatus, 
		INT4 NumberOfEntriesReturned, 
		INT4 PositionOfTerm,
		const CHR *ListEntries,
		const CHR *AttributeSet);
/*@Doc:   Construct a Scan Request/Response PDU from a parameter list*/

///Check for existance of Entry  
  inline INT HasRetStepSize   () const	
    { return c_root->HasTag(  (  3  )  );}; 	
/*@Doc: Return 1 if entry exists, 0 if it does not*/

///Get value of entry
  inline INT GetRetStepSize   (  INT4   *argValue) const	
    { return c_root->GetInt  (  (  3  )  ,argValue);}; 	
/*@Doc: Get the value of the entry in {\tt argValue} and return 1 on success, 0 on failure*/

///Set entry value
  inline INT AddRetStepSize   (  INT4   argValue) const	
    { return c_root->AddInt  (  (  3  )  ,argValue);};   
/*@Doc: Set the entry to the value stored in {\tt argValur} and return 1 on success, 0 on failure*/

 ///Check for existance of Entry 
  inline INT HasScanStatus   () const	
    { return c_root->HasTag(  (  4  )  );}; 	
/*@Doc: Return 1 if entry exists, 0 if it does not*/

///Get value of entry
  inline INT GetScanStatus   (  INT4   *argValue) const	
    { return c_root->GetInt  (  (  4  )  ,argValue);}; 	
/*@Doc: Get the value of the entry in {\tt argValue} and return 1 on success, 0 on failure*/

///Set entry value 	
  inline INT AddScanStatus   (  INT4   argValue) const	
    { return c_root->AddInt  (  (  4  )  ,argValue);};   
/*@Doc: Set the entry to the value stored in {\tt argValur} and return 1 on success, 0 on failure*/

///Check for existance of Entry
  inline INT HasNumberOfEntriesReturned   () const	
    { return c_root->HasTag(  (  5  )  );}; 	
/*@Doc: Return 1 if entry exists, 0 if it does not*/
 
///Get value of entry
  inline INT GetNumberOfEntriesReturned   (  INT4   *argValue) const	
    { return c_root->GetInt  (  (  5  )  ,argValue);}; 	
/*@Doc: Get the value of the entry in {\tt argValue} and return 1 on success, 0 on failure*/

///Set entry value 
  inline INT AddNumberOfEntriesReturned   (  INT4   argValue) const	
    { return c_root->AddInt  (  (  5  )  ,argValue);};   
/*@Doc: Set the entry to the value stored in {\tt argValur} and return 1 on success, 0 on failure*/

///Check for existance of Entry 
  inline INT HasPositionOfTerm   () const	
    { return c_root->HasTag(  (  6  )  );}; 	
/*@Doc: Return 1 if entry exists, 0 if it does not*/

///Get value of entry
  inline INT GetPositionOfTerm   (  INT4   *argValue) const	
    { return c_root->GetInt  (  (  6  )  ,argValue);}; 	
/*@Doc: Get the value of the entry in {\tt argValue} and return 1 on success, 0 on failure*/

///Set entry value 
  inline INT AddPositionOfTerm   (  INT4   argValue) const	
    { return c_root->AddInt  (  (  6  )  ,argValue);};   
/*@Doc: Set the entry to the value stored in {\tt argValur} and return 1 on success, 0 on failure*/  

///Check for existance of Entry
  inline INT HasListEntries  () const	
    { return c_root->HasTag(   7   );}; 	
/*@Doc: Return 1 if entry exists, 0 if it does not*/

///Add List Entries BERBROWSER
  inline INT AddListEntries () const	
    { return c_root->AddSubdirectory(  7  );};	
/*@Doc: Adds an empty List Entries BERBROWSER.  Returns 1 on success, 0 on failure*/

///Get List Entries BERBROWSER
  inline INT GetListEntries (PBERBROWSER & Dir) const	
    { return c_root->GetSubdirectory(  7  ,Dir);}; 
/*@Doc: Return the BERBROWSER associated with list entries.  Returns 1 on success, 0 on failure*/

///Check for existance of Entry
  inline INT HasAttributeSet   () const	
    { return c_root->HasTag(  (  8  )  );}; 	
/*@Doc: Return 1 if entry exists, 0 if it does not*/

///Get value of entry
  inline INT GetAttributeSet   (  STRING   *argValue) const	
    { return c_root->GetOID  (  (  8  )  ,argValue);}; 	
/*@Doc: Get the value of the entry in {\tt argValue} and return 1 on success, 0 on failure*/

///Set entry value 
  inline INT AddAttributeSet   (  STRING   argValue) const	
    { return c_root->AddOID  (  (  8  )  ,argValue);};   
  /*@Doc: Set the entry to the value stored in {\tt argValur} and return 1 on success, 0 on failure*/
};

//***************************************************************************
//
//	class ZSORTBASE - Base class for Z39.50 Sort PDUs
//	
//***************************************************************************

/// Main Z39.50 Sort Request/Response Base Class
/*@Doc:
\begin{description}

\item[Function]  Basic Sort PDU.  Sort Request/Response are derived from this base class.
\item[Algorithms] Basic tagged data insertion/retrieval into a tree.

\item[Flow] This is a fundamental c++ class.    Flow is described as needed within the individual method descriptions.
\item[Packaging] This fundamental class is linked and archived as determined by the compiling software.
\item[Security] This class has no security role or function.
\end{description}
*/
class ZSORTBASE : public ZPDU {
public:
///Basic Constructor
  ZSORTBASE(ASN1TAG Tag) : ZPDU(Tag) {}
/*@Doc: Build a Sort Request/Response from parameters (just the tag, in this case)*/
///Basic Buffer Constructor
  ZSORTBASE(PUCHR Buf, INT4 Length) : ZPDU(Buf,Length) {}
/*@Doc: Build a Sort Request from a buffer of BER encoded data*/
};


//***************************************************************************
//
//	class ZSORTREQUEST - Z39.50 Sort Request PDU
//	
//***************************************************************************


/// Z39.50 Sort Request PDU
/*@Doc:
\begin{description}

\item[Function]
\item[Algorithms]  The Sort Requests signals the target to sort a given result set on a particular element set, and name the new result set.

\item[Flow] This is a fundamental c++ class.    Flow is described as needed within the individual method descriptions.
\item[Packaging] This fundamental class is linked and archived as determined by the comiling software.
\item[Security] This class has no security role or function.
\end{description}
*/
class ZSORTREQUEST : public ZSORTBASE {
public:
///Basic Buffered Constructor
  ZSORTREQUEST(PUCHR Buf, INT4 Length) : ZSORTBASE(Buf,Length) {}
/*@Doc: Build Sort Request from a buffer of BER encoded data*/

///Basic Constructor
  ZSORTREQUEST(	const STRING& RefId,
	       const CHR *InputResultSetNames,
	       const STRING& SortedResultSetName,
	       const CHR *SortSequence);  
/*@Doc: Build a Sort request from parameters*/
  
///Check for entry
  inline INT HasInputResultSetNames  () const	
    { return c_root->HasTag(  3   );}; 	

///Add Entry	
  inline INT AddInputResultSetNames () const	
    { return c_root->AddSubdirectory( 3  );};	

///Get entry
  inline INT GetInputResultSetNames (PBERBROWSER & Dir) const	
    { return c_root->GetSubdirectory( 3  ,Dir);}; 

///Check for entry
  inline INT HasSortedResultSetName   () const	
    { return c_root->HasTag(  ( 4  )  );}; 	

///Get entry
  inline INT GetSortedResultSetName   (  STRING   *argValue) const	
    { return c_root->GetChar  (  ( 4  )  ,argValue);}; 	

///Add Entry	
  inline INT AddSortedResultSetName   (const   STRING   & argValue) const
    { return c_root->AddChar  (  ( 4  )  ,argValue);};   

///Check for entry
  inline INT HasSortSequence  () const	
    { return c_root->HasTag(  5   );}; 	

///Add Entry
  inline INT AddSortSequence () const	
    { return c_root->AddSubdirectory( 5  );};	

///Get entry
  inline INT GetSortSequence (PBERBROWSER & Dir) const	
    { return c_root->GetSubdirectory( 5  ,Dir);}; 
};



//***************************************************************************
//
//	class ZSORTRESPONSE - Z39.50 Sort Response PDU
//	
//***************************************************************************


/// Z39.50 Sort Response PDU Class
/*@Doc:
\begin{description}

\item[Function]  The Sort Response PDU returns the status of the requested
Sort operation.
\item[Algorithms]  Simple insertion/deletion of tree members

\item[Flow] This is a fundamental c++ class.    Flow is described as needed within the individual method descriptions.
\item[Packaging] This fundamental class is linked and archived as determined by the comiling software.
\item[Security] This class has no security role or function.
\end{description}
*/

class ZSORTRESPONSE : public ZSORTBASE {
public:
  /// Basic Buffer Constructor
    ZSORTRESPONSE(PUCHR Buf, INT4 Length) : ZSORTBASE(Buf,Length) {}
/*@Doc: Constructs a Sort Response from a buffer of data*/

    /// Basic Paramter List Constructor
      ZSORTRESPONSE(	const STRING& RefId,
		    const INT4 SortStatus,
		    const INT4 SortResponse,
		    const CHR *Diagnostics);
/*@Doc: Construct a Sort Response PDU from a paramete rlist */

    ///Checks for existence of Sort status entry
      inline INT HasSortStatus   () const	
	{ return c_root->HasTag(  (  3  )  );}; 	
/*@Doc: Return 1 if entry exists, 0 if not*/
    
    /// Gets Sort status entry
      inline INT GetSortStatus   (  INT4   *argValue) const	
	{ return c_root->GetInt  (  (  3  )  ,argValue);}; 	
/*@Doc Get the Sort Status code in {\tt argValue}*/
	
    /// Sets Sort status entry
      inline INT AddSortStatus   (  INT4   argValue) const	
	{ return c_root->AddInt  (  (  3  )  ,argValue);};
/*@Doc: Add {\tt argValue} as the Sort Status code*/

    /// Checks for Sort Response   
      inline INT HasSortResponse   () const	{ 
	return c_root->HasTag(  (  44  )  );}; 	
/*@Doc: Return 1 if entry exists, 0 if not*/

    /// Gets Sort Response
      inline INT GetSortResponse   (  INT4   *argValue) const	
	{ return c_root->GetInt  (  (  44  )  ,argValue);}; 	
/*@Doc: Get the Sort Response code.  Return 1 on success*/

    /// Adds Sort response
      inline INT AddSortResponse   (  INT4   argValue) const	
	{ return c_root->AddInt  (  (  44  )  ,argValue);};   
/*@Doc: Add {\tt argValue} as the Sort Response code */

    /// Checks for Sort sequence info
      inline INT HasSortSequence  () const	
	{ return c_root->HasTag(  5   );}; 	
	/*@Doc: Return 1 if entry exists, 0 if not*/

    /// Adds Sort sequence info
      inline INT AddSortSequence () const	
	{ return c_root->AddSubdirectory( 5  );};	
/*@Doc: Add an empty Sort Sequence BERBROWSER*/

    /// Server gets Sort sequence info
      inline INT GetSortSequence (PBERBROWSER & Dir) const	
	{ return c_root->GetSubdirectory( 5  ,Dir);}; 
/*@Doc: Get the Sort Sequence BERBROWSER*/
  };


//***************************************************************************
//
//	class ZESBASE - Base class for Z39.50 Extended Services PDUs
//	
//***************************************************************************
class ZESBASE : public ZPDU {
public:
	ZESBASE(ASN1TAG Tag) : ZPDU(Tag) {}
	ZESBASE(PUCHR Buf, INT4 Length) : ZPDU(Buf,Length) {}
};

//***************************************************************************
//
//	class ZESREQUEST - Z39.50 Extended Service Request PDU
//	
//***************************************************************************
class ZESREQUEST : public ZESBASE {
public:
	ZESREQUEST(PUCHR Buf, INT4 Length) : ZESBASE(Buf,Length) {}
	ZESREQUEST(	const STRING& RefId,
			const INT4 Function,
			const STRING& PackageType,
			const STRING& PackageName,
			const STRING& UserId,
			const INT4 RetentionTime,
			const STRING& Description,
			const INT4 WaitAction);
			
	DD_INT(Function, ES_FUNCTION_TAG)
	DD_OID(PackageType, ES_PACKAGETYPE_TAG)
	DD_STRING(PackageName, ES_PACKAGENAME_TAG)
	DD_STRING(UserId, ES_USERID_TAG)
	DD_INT(RetentionTime, ES_RETENTIONTIME_TAG)
	DD_SUBDIR(Permissions, ES_PERMISSIONS_TAG)
	DD_STRING(Description, ES_DESCRIPTION_TAG)
	DD_INT(WaitAction, ES_WAITACTION_TAG)
};

//***************************************************************************
//
//	class ZESRESPONSE - Z39.50 Extended Service Response PDU
//	
//***************************************************************************
class ZESRESPONSE : public ZESBASE {
public:
	ZESRESPONSE(PUCHR Buf, INT4 Length) : ZESBASE(Buf,Length) {}
	ZESRESPONSE(	const STRING& RefId,
			const INT4 OperationStatus,
			ZDEFAULTDIAGFORMAT *DiagRecord,
			ZESTASKPACKAGE *TaskPackage);

	DD_INT(OperationStatus, ES_OPERATIONSTATUS_TAG)
	DD_SUBDIR(Diagnostics, ES_DIAGNOSTICS_TAG)
	DD_SUBDIR(TaskPackage, ES_TASKPACKAGE_TAG)
	GDT_BOOLEAN GetTaskStatus(INT4 *TaskStatus);
};

//***************************************************************************
//
//	class ZESUPDATEREQUEST - Z39.50 ES Update Request PDU
//	
//***************************************************************************
class ZESUPDATEREQUEST : public ZESREQUEST {
public:
	ZESUPDATEREQUEST(PUCHR Buf, INT4 Length) : ZESREQUEST(Buf,Length) {}
	ZESUPDATEREQUEST(const STRING& RefId,
			const INT4 Action,
			const STRING& DatabaseName,
			ZRECORDLIST *Records);
	GDT_BOOLEAN GetAction(INT4 *Action);
	GDT_BOOLEAN GetDatabaseName(STRING *DatabaseName);
};

//***************************************************************************
//
//	class ZESUPDATERESPONSE - Z39.50 ES Update Response PDU
//	
//***************************************************************************
class ZESUPDATERESPONSE : public ZESRESPONSE {
public:
	ZESUPDATERESPONSE(PUCHR Buf, INT4 Length) : ZESRESPONSE(Buf,Length) {}
	ZESUPDATERESPONSE(const STRING& RefId,
			ZESTASKPACKAGE *TaskPackage,
			const INT4 Action,
			const STRING& DatabaseName,
			const INT4 UpdateStatus,
			ZDEFAULTDIAGFORMAT *Record);
	GDT_BOOLEAN GetUpdateStatus(INT4 *UpdateStatus);
};


//***************************************************************************
//
//	class ZCLOSE - Z39.50 Close PDU
//	
//***************************************************************************


/// Z39.50 Close PDU Class
/*@Doc:
\begin{description}

\item[Function] The Z39.50 Close service allows either an origin or target to abruptly terminate all active operations and to initiate termination of the current Z-association.
\item[Algorithms] Simple tree insertion/deletion

\item[Flow] This is a fundamental c++ class.    Flow is described as needed within the individual method descriptions.
\item[Packaging] This fundamental class is linked and archived as determined by the comiling software.
\item[Security] This class has no security role or function.
\end{description}
*/

class ZCLOSE : public ZPDU {
public:
  /// Basic Buffer Constructor
    ZCLOSE(PUCHR Buf, INT4 Length) : ZPDU(Buf, Length) {}
/*@Doc: Construct a Close object from buffer of BER encoded data.  Typically a server function to build Close object from buffer of network data*/

    /// Basic Constructor
      ZCLOSE(const CHR *RefId, INT4 Reason);
/*@Doc: Create a Close PDU object from parameters.  Typically a client function to build a Close object for encoding.*/
    
    /// Checks for existance of Close reason
      inline INT HasCloseReason   () const	
	{ return c_root->HasTag(  (  211  )  );}; 	
/*@Doc: Returns 1 if CloseReason is in PDU*/

    /// Gets reason for Close
      inline INT GetCloseReason   (  INT4   *argValue) const	
	{ return c_root->GetInt  (  (  211  )  ,argValue);}; 	
/*@Doc: Puts CloseReason in {\tt argValue}. Returns 1 on success. */	
    
    /// Adds reason for Close request
      inline INT AddCloseReason   (  INT4   argValue) const	
	{ return c_root->AddInt  (  (  211  )  ,argValue);};   
/*@Doc: Adds CloseReason to Close PDU*/  
  };

#endif
