/* $Id: zrecords.hxx,v 1.3 1998/01/30 19:26:41 cnidr Exp $ */
/************************************************************************
Copyright Notice

Copyright (c) MCNC, Clearinghouse for Networked Information Discovery and
Retrieval, 1994. 

Permission to use, copy, modify, distribute, and sell this software and
its documentation, in whole or in part, for any purpose is hereby granted
without fee, provided that

1. The above copyright notice and this permission notice appear in all
copies of the software and related documentation. Notices of copyright
and/or attribution which appear at the beginning of any file included in
this distribution must remain intact. 

2. Users of this software agree to make their best efforts (a) to return
to MCNC any improvements or extensions that they make, so that these may
be included in future releases; and (b) to inform MCNC/CNIDR of noteworthy
uses of this software. 

3. The names of MCNC and Clearinghouse for Networked Information Discovery
and Retrieval may not be used in any advertising or publicity relating to
the software without the specific, prior written permission of MCNC/CNIDR. 

THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY WARRANTY
OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE. 

IN NO EVENT SHALL MCNC/CNIDR BE LIABLE FOR ANY SPECIAL, INCIDENTAL,
INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES WHATSOEVER
RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT ADVISED OF THE
POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY, ARISING OUT OF OR
IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE. 
************************************************************************/

/*@@@
File:		zrecords.hxx
Version:	1.00
Description:	Z39.50 record classes
Author:		Kevin Gamiel, kevin.gamiel@cnidr.org
@@@*/

#ifndef _ZRECORDS_HXX_
#define _ZRECORDS_HXX_

#include <iostream.h>
#include "ber.hxx"
#include "ztags.hxx"


/// Base class for low-level Z39.50 records:
/*@Doc:
\begin{description}

\item[Function] Manage basic Z39.50 Records.  Note that this base class
has no data - derived virtual methods and  data must be added for real records.
\item[Algorithms]  None - basic class for all record functions. 

\item[Flow] This is a fundamental c++ class.    Flow is described as needed within the individual method descriptions.
\item[Packaging] This fundamental class is linked and archived as determined by the compiling software.
\item[Security] This class has no security role or function.
\end{description}
*/

class ZRECORDBASE {
public:
///Basic Constructor
	ZRECORDBASE();
/*@Doc: Basic record constructor - replaced by specific record-type 
constructor*/

///Basic Record Destructor
	virtual ~ZRECORDBASE();
/*@Basic delete function - replaced by specific functions in derived classes*/

///Add BER Node
	virtual void AddBER(BERBROWSER &ber);	
/*@Doc: Adds a ber node to the record*/

///Virtual Print Function
	virtual void Print(ostream & os) { cout << "*****" << endl;}
/*@Doc: Placeholder for derived record printing functions*/

///Get Record Data
	virtual void GetRecordData(STRING *Data) {}
/*@Doc: Derive a record extraction method*/

///Get Object Identifier	
	virtual void GetOID(STRING *oid) {}
/*@Doc Derive an OID extraction method*/

///Get Record Occurance
	virtual void GetRecordOcc(INT *Occ) {}
/*Doc: Get Occurance data from a record - unused*/
	
};


/// Base class for EXTERNAL records
/*@Doc:
\begin{description}

\item[Function] Manage basic external Z39.50 Records.  Note that this base 
class still has no data - derived virtual methods and  data must be added for 
external records.
\item[Algorithms]  None - basic class for all external record functions. 

\item[Flow] This is a fundamental c++ class.    Flow is described as needed within the individual method descriptions.
\item[Packaging] This fundamental class is linked and archived as determined by the compiling software.
\item[Security] This class has no security role or function.
\end{description}
*/

class ZEXTBASE : public ZRECORDBASE {
protected:
	GDT_BOOLEAN	c_is_retrieval_record;
	STRING 		c_oid;
public:
///Basic Constructor
	ZEXTBASE(const STRING & oid, const GDT_BOOLEAN IsRetrievalRecord=
		GDT_FALSE);	// See note in constructor code
/*Doc:   
     This is strange and I believe the result of an error in the Z39.50
     ASN.1.  In the NamePlusRecord SEQUENCE there is a tag record=1.
     Under that tag is a choice among several different record types.
     The first is retrievalRecord=1 which is of type EXTERNAL.  I believe
     the retrievalRecord tag is unecessary and is causing my C++ record
     class structure problems.  I need a standalone EXTERNAL record in other
     places in the standard so I need to derive it directly from 
     ZRECORDBASE, which imposes no tags of its own.  However, I also
     need an EXTERNAL hanging from a retrievalRecord (with tag of 1)
     for the case I've just described.  I may simply be missing something
     b/c I'm tired, but the only way I can fix this now is to add a flag
     to the ZEXTBASE class that says whether the particular instance of
     the EXTERNAL is really a retrievalRecord or not.  That way, the
     descendants can reset that flag appropriately and the AddBER()
     method for ZEXTBASE can either add the retrievalRecord tag or not.
     I realize this is an abuse of C++, but its not the first and surely
     won't be the last;-)
	*/
///Get Object Identifier
	void GetOID(STRING *oid);
/*Doc: Returns the OID*/

///Add BER Node
	virtual void AddBER(BERBROWSER &ber);	
/*@Doc: Adds external BER node if this is a retrieval record */

///Basic destructor
	virtual ~ZEXTBASE();
/*@Doc: Clean up this record */

///Print
	virtual void Print(ostream & os) { cout << "EXTBASE" << endl;}
/*@Doc:  Derived print function - no data yet, so prints record type*/
};


/// External Record, single-ASN1-type

class ZEXTSINGLE : public ZEXTBASE {
protected:
	STRING c_data;
public:
	ZEXTSINGLE(const STRING & oid, const STRING & data,
		const GDT_BOOLEAN IsRetrievalRecord=GDT_FALSE);
	virtual void Print(ostream & os) { cout << "ZEXTSINGLE:" << c_data; }
	virtual void AddBER(BERBROWSER &ber);	
};


/// External Record, octet-aligned

class ZEXTOCTETALIGNED : public ZEXTBASE {
protected:
	STRING c_data;
public:
	ZEXTOCTETALIGNED(const STRING & oid, const STRING & data,
		const GDT_BOOLEAN IsRetrievalRecord=GDT_FALSE);
	virtual void Print(ostream & os) { cout << "ZEXTOCTETALIGNED:" << c_data; }
	virtual void AddBER(BERBROWSER &ber);	
};


/// External Record, arbitrary choice

class ZEXTARBITRARY : public ZEXTBASE {
public:
	ZEXTARBITRARY();
};


/// SUTRS - Simple Unstructured Text Record Syntax

/*@Doc:	Encoded as an EXTERNAL, single-ASN1-encoding */
//
class ZSUTRSRECORD : public ZEXTSINGLE {
public:
	ZSUTRSRECORD(const STRING & Data);
	ZSUTRSRECORD(const BERBROWSER &ber);
	virtual void GetRecordData(STRING *Data) { *Data = c_data; }
	virtual void Print(ostream & os);
};


/// USMARC Record Syntax

//@Doc:	Encoded as an EXTERNAL, octet-aligned
//
class ZUSMARCRECORD : public ZEXTOCTETALIGNED {
public:
	ZUSMARCRECORD(const STRING & Data);
	virtual void GetRecordData(STRING *Data) { *Data = c_data; }
};

//
// Extended services task package
//
class ZESTASKPACKAGE : public ZEXTBASE {
protected:
	STRING	c_package_type;
	STRING	c_package_name;
	STRING	c_user_id;
	INT4	c_retention_time;
	STRING	c_description;
	INT4	c_task_status;
public:
	ZESTASKPACKAGE(const STRING & package_type,
		const STRING & package_name,
		const STRING & user_id,
		const INT4 & retention_time,
		const STRING & description,
		const INT4 & task_status);
	virtual void Print(ostream & os);
	virtual void AddBER(BERBROWSER &ber);	
};


/// Virtual base class for Records CHOICE

class ZRECORDSCHOICE {
 public:
  ZRECORDSCHOICE();
  virtual ~ZRECORDSCHOICE();
	
  virtual void AddBER(BERBROWSER &ber);	
  virtual void Print(ostream & os) {cout << "***" << endl;}
  virtual void GetRecordData(STRING *Data) { }
  virtual void GetRecordOcc(INT *Occ){};
  virtual void GetOID(STRING *oid) {}
virtual void GetRecordCondition(INT *Condition){}
};


// NAMEPLUSRECORD Encoding Structure

class ZNAMEPLUSRECORD : public ZRECORDSCHOICE {
private:
	STRING	c_db_name;
	ZRECORDBASE *c_record;
public:
	ZNAMEPLUSRECORD(const STRING & DatabaseName, 
		ZRECORDBASE *Record);
	~ZNAMEPLUSRECORD();

	virtual void AddBER(BERBROWSER &ber);	
	virtual void Print(ostream & os);
	virtual void GetRecordOcc(INT *Occ);
	virtual void GetRecordData(STRING *Data) { c_record->GetRecordData(Data); }
	virtual void GetOID(STRING *oid) { c_record->GetOID(oid); }
};

//
// Extended service update record 
//
class ZESUPDATERECORD : public ZRECORDSCHOICE {
private:
	STRING	c_record_id;
	ZRECORDBASE *c_record;
public:
	ZESUPDATERECORD(const STRING & RecordId, 
		ZRECORDBASE *Record);
	~ZESUPDATERECORD();

	virtual void AddBER(BERBROWSER &ber);	
	virtual void Print(ostream & os);
	virtual void GetRecordData(STRING *Data) { c_record->GetRecordData(Data); }
	virtual void GetOID(STRING *oid) { c_record->GetOID(oid); }
};

//
/// Default diagnostic record.  
//
/*@Doc: The AddInfo is either a Visible String or an International String 
 depending on the Z39.50 version.  Instead
 of making separate classes, I've added a flag in the constructor
 that defaults to version 2.  Simply specify version 3 or above
 to get an International String.
*/
class ZDEFAULTDIAGFORMAT : public ZRECORDSCHOICE {
	STRING	c_oid,
		c_addinfo;
	INT4	c_condition,
		c_version;
public:
	ZDEFAULTDIAGFORMAT(const INT4 Condition, 
		const STRING & AddInfo, 
		const STRING & DiagSetId,
		const INT Version=2);
	void AddBER(BERBROWSER &ber);
        void GetRecordData(STRING *Data) { *Data=c_addinfo;}
        void GetRecordCondition(INT *Condition){*Condition=c_condition;}
};

/// A List of returned records
class ZRECORDLIST {
	ZRECORDSCHOICE **c_recordlist;
	INT 	c_maxrecords,
		c_recordcount;
	ASN1TAG	*c_recordtype;
	CHR 	**c_dbname;
	void 	Init(INT Tag);
public:
///Create the List
	ZRECORDLIST();
///Create the List
	ZRECORDLIST(INT Tag);
///Destroy the list
	virtual ~ZRECORDLIST();
///Add A Record and DBName to List
	virtual void AddRecord(ZRECORDSCHOICE *Record, CHR *dbn);
///Add A Record to List
	virtual void AddRecord(ZRECORDSCHOICE *Record);
///Get a Record
	virtual ZRECORDSCHOICE *GetRecord(const INT RecNum);
///Get a Record and DB Name
	virtual ZRECORDSCHOICE *GetRecord(const INT RecNum, CHR *dbn);
///Get Count of Records
	virtual const INT GetRecordCount() { return c_recordcount; }
///Adds BER Node to Each Record
	virtual void AddBER(BERBROWSER &ber);
/*@Doc:Given a PRESENTRESPONSE PDU, for example, adds every BER node from
the RECORDS node on down.*/

///Print the record
	virtual void Print(ostream & os);
///Set the Type for the list
	virtual void SetRecordType(INT Tag);
///Clear the list
	void Clear();	
};
/*@Doc: provides a mechanism for manipulating a list of records*/


/// A set of response records
class ZRESPONSERECORDS : public ZRECORDLIST {
public:
	ZRESPONSERECORDS();
};
/*@Doc: Create a list of records to be returned in a response */

/// Convenience function.  Builds a record list consisting of a single nonsurrogate diagnostic record of default diag format.

ZRECORDLIST *NonsurrogateDiagnosticRecord(const INT4 Condition,
        const STRING & AddInfo, const STRING & DiagSetId,
        const INT Version=2);
/*@Doc: Builds a record list consisting of a single nonsurrogate diagnostic record of default diag format.*/


/// An Entry in a record
class ZENTRY : public ZRECORDBASE {
protected:
        STRING c_data;
        INT    c_occ;
public:
///Basic Constructor
        ZENTRY(const STRING & data, const INT & occ);
/*@Doc: Create a basic entry in a record*/

///Basic Destructor
        virtual ~ZENTRY();
/*@Doc: Delete an entry in a record*/
///Get record data
        virtual void GetRecordData(STRING *Data) { *Data = c_data; }
/*@Doc: Retrieve the data from the entry*/
///Get occurrance count
        virtual void GetRecordOcc(INT *Occ) { *Occ = c_occ; }
};          



#endif
