/* $Id: zsession.hxx,v 1.5 1997/09/28 16:28:33 cnidr Exp $ */
/************************************************************************
Copyright Notice

Copyright (c) MCNC, Clearinghouse for Networked Information Discovery and
Retrieval, 1994. 

Permission to use, copy, modify, distribute, and sell this software and
its documentation, in whole or in part, for any purpose is hereby granted
without fee, provided that

1. The above copyright notice and this permission notice appear in all
copies of the software and related documentation. Notices of copyright
and/or attribution which appear at the beginning of any file included in
this distribution must remain intact. 

2. Users of this software agree to make their best efforts (a) to return
to MCNC any improvements or extensions that they make, so that these may
be included in future releases; and (b) to inform MCNC/CNIDR of noteworthy
uses of this software. 

3. The names of MCNC and Clearinghouse for Networked Information Discovery
and Retrieval may not be used in any advertising or publicity relating to
the software without the specific, prior written permission of MCNC/CNIDR. 

THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY WARRANTY
OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE. 

IN NO EVENT SHALL MCNC/CNIDR BE LIABLE FOR ANY SPECIAL, INCIDENTAL,
INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES WHATSOEVER
RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT ADVISED OF THE
POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY, ARISING OUT OF OR
IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE. 
************************************************************************/

/*@@@
File:		zsession.hxx
Version:	1.00
Description:	Z39.50 Session.  Base class for ZCLIENT and ZSERVER.
Author:		Kevin Gamiel, kevin.gamiel@cnidr.org
@@@*/

#ifndef _ZSESSION_H_
#define _ZSESSION_H_

//#include <values.h>	// for MAXINT
#include <limits.h>	// for INT_MAX
#include "tcpsock.hxx"
#include "zpdu.hxx"
#include "registry.hxx"

#define ZPREFMSGSIZE 32768
#define ZMAXRECORDSIZE 8388608
#define ZPROTVER "yyy"
#define ZOPTIONS "yynnnnnyynnnnnn"
#define CNIDR_IMP_ID "34"

/// Main Z39.50 Session Manager Class
/*@Doc:
\begin{description}  

\item[Function] This completely manages a Z39.50 Session (Client or Server)
\item[Algorithms]  This class provides access to appropriate client and server classes to fully support a Z39.50 session.

\item[Flow] This is a fundamental c++ class.    Flow is described as needed within the individual method descriptions.
\item[Packaging] This fundamental class is linked and archived as determined by the compiling software.
\item[Security] This class has no security role or function.
\end{description}
*/

class ZSESSION {
protected:
	TCPSOCK 	*c_tcp;	// network connection
	INT 		c_debuglevel;
	enum 		{ open, closed } c_state;
	INT4 		c_prefmsgsize,
			c_maxrecordsize,
			c_timeout;
	UINT		c_port;
	STRING		c_protocol_version,
			c_options,
			c_refid,
			c_group,
			c_inifile,
			c_dbname;
///Read a Z39.50 PDU
	PUCHR 		ReadPDU(INT4 *len, unsigned short *type);
///Read a Present PDU
	PUCHR 		ReadPresentPDU(INT4 *len, unsigned short *type);
///Send PDU over Network
	INT4 		SendPDU(PZPDU pdu);
///Send a Present PDU over the network
	INT4 		SendPresentPDU(PZPDU pdu);
	REGISTRY	*c_defaults;
///Load Session Defaults
	void LoadDefaultsFromFile(const STRING & Filename,const STRING & Group);
///Load Session Defaults
	void LoadDefaultsFromCommandLine(int ac, char **argv, STRING & File,
        	STRING & Group, STRING *LastGroup, STRING *LastFile);

///Set Session protocol version Option
	void SetProtocolVersion(const STRING & Value) 
		{ c_protocol_version = Value; }

///Set Session Message Size Option
	void SetPreferredMessageSize(const INT4 Value) 
		{ c_prefmsgsize = Value; }
///Set Maximum Message Size Option
	void SetExceptionalRecordSize(const INT4 Value) 
		{ c_maxrecordsize = Value; }

	//
	// Options
	//
///Clear Session Options
	void ClearOptions();
///Set Session Option
	void SetOption(const INT Option, const GDT_BOOLEAN Value);
///Set Session Option
	void SetOptions(const STRING & Options);

	enum {	SEARCH,
		PRESENT,
		DELSET,
		RESOURCEREPORT,
		TRIGGERRESOURCECTRL,
		RESOURCECTRL,
		ACCESSCTRL,
		SCAN,
		SORT,
		RESERVED,
		EXTENDEDSERVICES,
		LEVEL1SEG,
		LEVEL2SEG,
		CONCURRENTOPS,
		NAMEDRESULTSETS };
	//
	// Misc
	//
///Print Connection Information
	void PrintConnectionDetails();

public:
///Create a Z39.50 Session
	ZSESSION();
/*@Doc: Create a basic Z39.50 session, from which clients and/or servers are derived*/

///Create a Z39.50 Session
	ZSESSION(const ZSESSION &OtherSession);
/*@Doc: Create a basic Z39.50 session, from which clients and/or servers are derived*/

///Destroy This Session
	~ZSESSION();
/*@Doc: Delete, close, and clean up this session.*/

	//
	// Shared by clients and servers
	//
///Close the session
	INT Close();
/*@Doc: Closes the network connection associated with this session*/

///Set Debug Level
	void SetDebugLevel(INT Level) { c_debuglevel = Level; }
/*@Doc: Set the debugging level for this session.  0 is lowest level, 100 is highest.  Higher numbers yield more verbose debugging messages*/

///Get Debug Level
	INT  GetDebugLevel() { return c_debuglevel; }
/*@Doc: Get the debugging level for this session.*/

///Get Option
	GDT_BOOLEAN GetOption(const INT Option);
///Get Protocol Version
	void GetProtocolVersion(INT *Version);
/*@Doc Get the Z39.50 Protocol Version for negotiated this session*/
};

#endif
